'''
whitelist_methods.py - Methods for whitelisting cell barcodes
=============================================================

:Author: Tom Smith, Ian Sudbery
:Release: $Id$
:Date: |today|
:Tags: Python UMI

'''
import collections
import re
import random
from functools import partial
# import umi_tools.Utilities as U


def get_barcode_read_id(read, cell_barcode=False, sep="_"):
    ''' extract the umi +/- cell barcode from the read id using the
    specified separator '''

    try:
        if cell_barcode:
            umi = read.qname.split(sep)[-1].encode('utf-8')
            cell = read.qname.split(sep)[-2].encode('utf-8')
        else:
            umi = read.qname.split(sep)[-1].encode('utf-8')
            cell = None

        return umi, cell

    except:
        raise ValueError(
            "Could not extract UMI +/- cell barcode from the read"
            "ID, please check UMI is encoded in the read name")


def get_barcode_tag(read,
                    cell_barcode=False,
                    umi_tag='RX',
                    cell_tag=None,
                    umi_tag_split=None,
                    umi_tag_delim=None,
                    cell_tag_split="-",
                    cell_tag_delim=None):
    ''' extract the umi +/- cell barcode from the specified tag '''

    try:
        if cell_barcode:
            umi = read.get_tag(umi_tag)
            cell = read.get_tag(cell_tag)
        else:
            umi = read.get_tag(umi_tag)
            cell = None

        if umi_tag_split:
            umi = umi.split(umi_tag_split)[0]

        if umi_tag_delim:
            umi = "".join(umi.split(umi_tag_delim))

        # 10X pipelines append a 'GEM' tag to the UMI, e.g
        # GATAGATACCTAGATA-1, hence default to split by "-
        if cell and cell_tag_split:
            cell = cell.split(cell_tag_split)[0]

        if cell and cell_tag_delim:
            cell = "".join(cell.split(cell_tag_delim))

        if cell:
            cell = cell.encode('utf-8')

        return umi.encode('utf-8'), cell

    except IndexError:
        raise ValueError("Could not extract UMI +/- cell barcode from the "
                         "read tag")


def get_umi_read_string(read_id, sep="_"):
    ''' extract the umi from the read id (input as a string) using the
    specified separator '''

    try:
        return (None, read_id.split(sep)[-1].encode('utf-8'))
    except IndexError:
        raise ValueError(
            "Could not extract UMI from the read ID, please"
            "check UMI is encoded in the read name")


def get_cell_umi_read_string(read_id, sep="_"):
    ''' extract the umi and cell barcode from the read id (input as a
    string) using the specified separator '''

    try:
        return (read_id.split(sep)[-1].encode('utf-8'),
                read_id.split(sep)[-2].encode('utf-8'))
    except IndexError:
        raise ValueError(
            "Could not extract UMI or CB from the read ID, please"
            "check UMI and CB are encoded in the read name:"
            "%s" % read_id)


def get_barcode_umis(read, cell_barcode=False):
    ''' extract the umi +/- cell barcode from the read name where the barcodes
    were extracted using umis'''

    umi, cell = None, None
    try:
        read_name_elements = read.qname.split(":")
        for element in read_name_elements:
            if element.startswith("UMI_"):
                umi = element[4:].encode('utf-8')
            elif element.startswith("CELL_") and cell_barcode:
                cell = element[5:].encode('utf-8')

        if umi is None:
            raise ValueError()

        return umi, cell

    except:
        raise ValueError("Could not extract UMI +/- cell barcode from the "
                         "read tag")


class get_bundles:

    ''' A functor - When called returns a dictionary of dictionaries,
    representing the unique reads at a position/spliced/strand
    combination. The key to the top level dictionary is a umi. Each
    dictionary contains a "read" entry with the best read, and a count
    entry with the number of reads with that
    position/spliced/strand/umi combination

    initiation arguments:

    options: script options

    all_reads: if true, return all reads in the dictionary. Else,
    return the 'best' read (using MAPQ +/- multimapping) for each key

    return_read2: Return read2s immediately as a single read

    metacontig_contig: Maps metacontigs to the consistuent contigs
    '''

    def __init__(self,
                 options,
                 only_count_reads=False,
                 all_reads=False,
                 return_unmapped=False,
                 return_read2=False,
                 metacontig_contig=None):

        self.options = options
        self.only_count_reads = only_count_reads
        self.all_reads = all_reads
        self.return_unmapped = return_unmapped
        self.return_read2 = return_read2
        self.metacontig_contig = metacontig_contig

        self.contig_metacontig = {}
        if self.metacontig_contig:
            for metacontig in metacontig_contig:
                for contig in metacontig_contig[metacontig]:
                    self.contig_metacontig[contig] = metacontig

        # set the method with which to extract umis from reads
        if self.options['get_umi_method'] == "read_id":
            self.barcode_getter = partial(
                get_barcode_read_id,
                cell_barcode=self.options['per_cell'],
                sep=self.options['umi_sep'])

        elif self.options['get_umi_method'] == "tag":
            self.barcode_getter = partial(
                get_barcode_tag,
                umi_tag=self.options['umi_tag'],
                cell_barcode=self.options['per_cell'],
                cell_tag=self.options['cell_tag'],
                umi_tag_split=self.options['umi_tag_split'],
                umi_tag_delim=self.options['umi_tag_delim'],
                cell_tag_split=self.options['cell_tag_split'],
                cell_tag_delim=self.options['cell_tag_delim'])

        elif self.options['get_umi_method'] == "umis":
            self.barcode_getter = partial(
                get_barcode_umis,
                cell_barcode=self.options['per_cell'])

        else:
            raise ValueError("Unknown UMI extraction method")

        self.read_events = collections.Counter()
        self.observed_contigs = collections.defaultdict(set)

        self.last_pos = 0
        self.last_chr = None
        self.start = 0
        self.current_chr = None

        self.reads_dict = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(dict)))
        self.read_counts = collections.defaultdict(
            lambda: collections.defaultdict(dict))

    def update_dicts(self, read, pos, key, umi):

        # The content of the reads_dict depends on whether all reads
        # are being retained
        if self.all_reads:

            # retain all reads per key
            try:
                self.reads_dict[pos][key][umi]["count"] += 1
            except KeyError:
                self.reads_dict[pos][key][umi]["read"] = [read]
                self.reads_dict[pos][key][umi]["count"] = 1
            else:
                self.reads_dict[pos][key][umi]["read"].append(read)

        elif self.only_count_reads:
            # retain all reads per key
            try:
                self.reads_dict[pos][key][umi]["count"] += 1
            except KeyError:
                self.reads_dict[pos][key][umi]["count"] = 1

        else:

            # retain just a single read per key
            try:
                self.reads_dict[pos][key][umi]["count"] += 1
            except KeyError:
                self.reads_dict[pos][key][umi]["read"] = read
                self.reads_dict[pos][key][umi]["count"] = 1
                self.read_counts[pos][key][umi] = 0
            else:
                if self.reads_dict[pos][key][umi]["read"].mapq > read.mapq:
                    return

                if self.reads_dict[pos][key][umi]["read"].mapq < read.mapq:
                    self.reads_dict[pos][key][umi]["read"] = read
                    self.read_counts[pos][key][umi] = 0
                    return

                # TS: implemented different checks for multimapping here
                if self.options['detection_method'] in ["NH", "X0"]:
                    tag = self.options['detection_method']
                    if (self.reads_dict[pos][key][umi]["read"].opt(tag) <
                        read.opt(tag)):
                        return
                    elif (self.reads_dict[pos][key][umi]["read"].opt(tag) >
                          read.opt(tag)):
                        self.reads_dict[pos][key][umi]["read"] = read
                        self.read_counts[pos][key][umi] = 0

                elif self.options['detection_method'] == "XT":
                    if self.reads_dict[pos][key][umi]["read"].opt("XT") == "U":
                        return
                    elif read.opt("XT") == "U":
                        self.reads_dict[pos][key][umi]["read"] = read
                        self.read_counts[pos][key][umi] = 0

                self.read_counts[pos][key][umi] += 1
                prob = 1.0/self.read_counts[pos][key][umi]

                if random.random() < prob:
                    self.reads_dict[pos][key][umi]["read"] = read

    def check_output(self):

        do_output = False
        out_keys = None
        c = 0

        if self.options['per_gene']:

            if self.metacontig_contig:

                if (self.current_chr != self.last_chr and
                    (self.observed_contigs[self.last_pos] ==
                     self.metacontig_contig[self.last_pos])):
                    do_output = True
                    out_keys = [self.last_pos]

            else:
                if self.current_chr != self.last_chr:
                    do_output = True
                    out_keys = sorted(self.reads_dict.keys())

        elif self.options['whole_contig']:

            if self.current_chr != self.last_chr:
                do_output = True
                out_keys = sorted(self.reads_dict.keys())
        else:
            if (self.start > (self.last_pos+1000) or
                self.current_chr != self.last_chr):
                # print(self.start, (self.last_pos), self.current_chr, self.last_chr)

                do_output = True
                out_keys = sorted(self.reads_dict.keys())

                if self.current_chr == self.last_chr:
                    out_keys = [x for x in out_keys if x <= self.start-1000]
            # else:
            #     print(self.start, (self.last_pos), self.current_chr, self.last_chr)

        return do_output, out_keys

    def __call__(self, inreads):
        c = 0
        for read in inreads:

            if read.is_read2:
                if self.return_read2:
                    if not read.is_unmapped or (
                            read.is_unmapped and self.return_unmapped):
                        yield read, None, "single_read"
                continue
            else:
                self.read_events['Input Reads'] += 1

            # only ever dealing with read1s from here

            if self.options['paired']:
                if read.is_paired:
                    self.read_events['Read pairs'] += 1
                else:
                    self.read_events['Unpaired reads'] += 1

                    # if paired end input and read1 is unpaired...

                    # skip, or
                    if self.options['unpaired_reads'] == "discard":
                        continue

                    # yield without grouping, or
                    elif self.options['unpaired_reads'] == "output":
                        yield read, None, "single_read"

                    # Use read pair; TLEN will be 0
                    elif self.options['unpaired_reads'] == "use":
                        pass

            if read.is_unmapped:
                if self.options['paired']:
                    if read.mate_is_unmapped:
                        self.read_events['Both unmapped'] += 1
                    else:
                        self.read_events['Read 1 unmapped'] += 1
                else:
                    self.read_events['Single end unmapped'] += 1

                # if read1 is unmapped, yield immediately or skip read
                if self.return_unmapped:
                    self.read_events['Input Reads'] += 1
                    yield read, None, "single_read"
                continue

            if self.options['paired'] and read.mate_is_unmapped:
                if not read.is_unmapped:
                    self.read_events['Read 2 unmapped'] += 1

                # if paired end input and read2 is unmapped, skip unless
                # options.unmapped_reads == "use", in which case TLEN will be 0
                if self.options['unmapped_reads'] != "use":
                    if self.return_unmapped:
                        yield read, None, "single_read"
                    continue

            if read.is_paired and (
                    read.reference_name != read.next_reference_name):
                self.read_events['Chimeric read pair'] += 1

                # if paired end input and read2 is mapped to another contig...

                # skip, or
                if self.options['chimeric_pairs'] == "discard":
                    continue

                # yield without grouping, or
                elif self.options['chimeric_pairs'] == "output":
                    yield read, None, "single_read"
                    continue

                # Use read pair; TLEN will be 0
                elif self.options['chimeric_pairs'] == "use":
                    pass

            if self.options['subset']:
                if random.random() >= self.options['subset']:
                    self.read_events['Randomly excluded'] += 1
                    continue

            if self.options['mapping_quality']:
                if read.mapq < self.options['mapping_quality']:
                    self.read_events['< MAPQ threshold'] += 1
                    continue

            # get the umi +/- cell barcodes
            if self.options['ignore_umi']:
                if self.options['per_cell']:
                    umi, cell = self.barcode_getter(read)
                    umi = ""
                else:
                    umi, cell = "", ""
            else:
                try:
                    umi, cell = self.barcode_getter(read)
                except KeyError:
                    error_msg = "Read skipped, missing umi and/or cell tag"
                    if self.read_events[error_msg] == 0:

                        # pysam renamed .tostring -> to_string in 0.14
                        # .tostring requies access to the parent AlignmentFile
                        try:
                            formatted_read = read.to_string()
                        except AttributeError:
                            formatted_read = read.query_name

                        U.warn("At least one read is missing UMI and/or "
                               "cell tag(s): %s" % formatted_read)
                    self.read_events[error_msg] += 1
                    continue

            self.current_chr = read.reference_name

            if self.options['per_gene']:

                if self.options['per_contig']:

                    if self.metacontig_contig:
                        transcript = read.reference_name
                        gene = self.contig_metacontig[transcript]
                    else:
                        gene = read.reference_name

                elif self.options['gene_tag']:

                    try:
                        assigned = read.get_tag(self.options['assigned_tag'])
                        gene = read.get_tag(self.options['gene_tag'])
                    except KeyError:
                        self.read_events['Read skipped, no tag'] += 1
                        continue

                    if gene == "":
                        if self.read_events['Read skipped - gene string is empty'] == 0:
                            U.warn("Assigned gene is empty string. First such "
                                   "read:\n%s" % read.to_string())
                        self.read_events['Read skipped - gene string is empty'] += 1
                        continue

                    if re.search(self.options['skip_regex'], assigned):
                        self.read_events['Read skipped - assigned tag matches skip_regex'] += 1
                        continue

                pos = gene
                key = pos

                if self.last_chr:
                    do_output, out_keys = self.check_output()
                else:
                    do_output = False

                if do_output:
                    for p in out_keys:
                        for k in sorted(self.reads_dict[p].keys()):
                            yield self.reads_dict[p][k], k, "bundle"

                        del self.reads_dict[p]

                self.last_chr = self.current_chr
                self.last_pos = pos

            else:
                self.start, pos, is_spliced = get_read_position(
                    read, self.options['soft_clip_threshold'])

                do_output, out_keys = self.check_output()
                if do_output:
                    c += 1
                    for p in out_keys:
                        # print(out_keys)
                        for k in sorted(self.reads_dict[p].keys()):
                            yield self.reads_dict[p][k], k, "bundle"

                        del self.reads_dict[p]
                        if p in self.read_counts:
                            del self.read_counts[p]
                # else:

                    # print(do_output)

                self.last_pos = self.start
                self.last_chr = self.current_chr

                if self.options['read_length']:
                    r_length = read.query_length
                else:
                    r_length = 0

                key = (read.is_reverse, self.options['spliced'] and is_spliced,
                       (not self.options['ignore_tlen'])*self.options['paired']*read.tlen, r_length)

            # update dictionaries
            key = (key, cell)
            self.update_dicts(read, pos, key, umi)

            if self.metacontig_contig:
                # keep track of observed contigs for each gene
                self.observed_contigs[gene].add(transcript)
        # yield remaining bundles
        print('tsdwwww', c)
        for p in sorted(self.reads_dict.keys()):
            for k in sorted(self.reads_dict[p].keys()):
                yield self.reads_dict[p][k], k, "bundle"


def get_gene_count_tab(infile,
                       bc_getter=None):

    ''' Yields the counts per umi for each gene

    bc_getter: method to get umi (plus optionally, cell barcode) from
    read, e.g get_umi_read_id or get_umi_tag


    TODO: ADD FOLLOWING OPTION

    skip_regex: skip genes matching this regex. Useful to ignore
                unassigned reads (as per get_bundles class above)

    '''

    gene = None
    counts = collections.Counter()

    for line in infile:

        values = line.strip().split("\t")

        assert len(values) == 2, "line: %s does not contain 2 columns" % line

        read_id, assigned_gene = values

        if assigned_gene != gene:
            if gene:
                yield gene, counts

            gene = assigned_gene
            counts = collections.defaultdict(collections.Counter)

        cell, umi = bc_getter(read_id)
        counts[cell][umi] += 1

    # yield final values
    yield gene, counts


class TwoPassPairWriter:
    '''This class makes a note of reads that need their pair outputting
    before outputting.  When the chromosome changes, the reads on that
    chromosome are read again, and any mates of reads already output
    are written and removed from the list of mates to output. When
    close is called, this is performed for the last chormosome, and
    then an algorithm identicate to pysam's mate() function is used to
    retrieve any remaining mates.

    This means that if close() is not called, at least as contigs
    worth of mates will be missing. '''

    def __init__(self, infile, outfile, tags=False):
        self.infile = infile
        self.outfile = outfile
        self.read1s = set()
        self.chrom = None

    def write(self, read, unique_id=None, umi=None, unmapped=False):
        '''Check if chromosome has changed since last time. If it has, scan
        for mates. Write the read to outfile and save the identity for paired
        end retrieval'''

        if unmapped or read.mate_is_unmapped:
            self.outfile.write(read)
            return

        if not self.chrom == read.reference_name:
            self.write_mates()
            self.chrom = read.reference_name

        key = read.query_name, read.next_reference_name, read.next_reference_start
        self.read1s.add(key)

        self.outfile.write(read)

    def write_mates(self):
        '''Scan the current chromosome for matches to any of the reads stored
        in the read1s buffer'''
        if self.chrom is not None:
            U.debug("Dumping %i mates for contig %s" % (
                len(self.read1s), self.chrom))

        for read in self.infile.fetch(reference=self.chrom, multiple_iterators=True):
            if any((read.is_unmapped, read.mate_is_unmapped, read.is_read1)):
                continue

            key = read.query_name, read.reference_name, read.reference_start
            if key in self.read1s:
                self.outfile.write(read)
                self.read1s.remove(key)

        U.debug("%i mates remaining" % len(self.read1s))

    def close(self):
        '''Write mates for remaining chromsome. Search for matches to any
        unmatched reads'''

        self.write_mates()
        U.info("Searching for mates for %i unmatched alignments" %
               len(self.read1s))

        found = 0
        for read in self.infile.fetch(until_eof=True, multiple_iterators=True):

            if any((read.is_unmapped, read.mate_is_unmapped, read.is_read1)):
                continue

            key = read.query_name, read.reference_name, read.reference_start
            if key in self.read1s:
                self.outfile.write(read)
                self.read1s.remove(key)
                found += 1
                continue

        U.info("%i mates never found" % len(self.read1s))
        self.outfile.close()


def getMetaContig2contig(bamfile, gene_transcript_map):
    ''' '''
    references = set(bamfile.references)
    metacontig2contig = collections.defaultdict(set)
    for line in U.openFile(gene_transcript_map, "r"):

        if line.startswith("#"):
            continue

        if len(line.strip()) == 0:
            break

        gene, transcript = line.strip().split("\t")
        if transcript in references:
            metacontig2contig[gene].add(transcript)

    return metacontig2contig


def metafetcher(bamfile, metacontig2contig, metatag):
    ''' return reads in order of metacontigs'''
    for metacontig in metacontig2contig:
        for contig in metacontig2contig[metacontig]:
            for read in bamfile.fetch(contig):
                read.set_tag(metatag, metacontig)
                yield read


def find_splice(cigar):
    '''Takes a cigar string and finds the first splice position as
    an offset from the start. To find the 5' end (read coords) of
    the junction for a reverse read, pass in the reversed cigar tuple'''

    offset = 0
    # a soft clip at the end of the read is taken as splicing
    # where as a soft clip at the start is not.
    if cigar[0][0] == 4:
        offset = cigar[0][1]
        cigar = cigar[1:]

    for op, bases in cigar:
        if op in (3, 4):
            # N or S: found the splice
            return offset
        elif op in (0, 2, 7, 8):
            # M, D, = or X: reference consuming
            offset += bases
        elif op in (1, 5, 6):
            # I, H, P: non-reference consuming
            continue
        else:
            raise ValueError("Bad Cigar operation: %i" % op)

    return False


def get_read_position(read, soft_clip_threshold):
    ''' get the read position (taking account of clipping) '''
    is_spliced = False

    if read.is_reverse:

        pos = read.aend
        if read.cigar[-1][0] == 4:

            pos = pos + read.cigar[-1][1]
        start = read.pos

        if ('N' in read.cigarstring or
            (read.cigar[0][0] == 4 and
             read.cigar[0][1] > soft_clip_threshold)):
            cigar = read.cigar[::-1]
            is_spliced = find_splice(cigar)
    else:
        pos = read.pos
        if read.cigar[0][0] == 4:
            pos = pos - read.cigar[0][1]
        start = pos

        if ('N' in read.cigarstring or
            (read.cigar[-1][0] == 4 and
             read.cigar[-1][1] > soft_clip_threshold)):
            is_spliced = find_splice(read.cigar)

    return start, pos, is_spliced



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import seaborn as sns
import matplotlib.pyplot as plt


class TraceSingle:

    def __init__(
            self,
            df_apv,
            df_disapv,
    ):
        self.df_apv = df_apv
        self.df_disapv = df_disapv
        self.method = self.df_disapv['method'].unique()[0]
        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def line_apv_disapv(self, ):
        fig, ax = plt.subplots(2, 1, figsize=(5, 6), sharex=True)
        colors = [
            'dimgray',  # black dimgray
            'palevioletred', # crimson mediumvioletred
        ]
        # colors = sns.color_palette("Set2")
        labels = [
            'different origin',
            'same origin',
        ]
        for i, col in enumerate(['diff_origin', 'same_origin']):
            ax[0].plot(
                self.df_disapv.index,
                self.df_disapv[col],
                label=labels[i],
                color=colors[i],
                lw=2.5,
                # alpha=0.8,
            )
            ax[1].plot(
                self.df_apv.index,
                self.df_apv[col],
                label=labels[i],
                color=colors[i],
                lw=2.5,
                # alpha=0.8,
            )

        ax[0].set_ylabel('UMI count', fontsize=14)
        ax[0].set_title('Not merged' + ' (' + self.method + ')', fontsize=12)
        ax[0].spines['right'].set_visible(False)
        ax[0].spines['top'].set_visible(False)

        ax[1].set_xlabel('PCR cycle', fontsize=12)
        ax[1].set_xticks(self.df_apv.index)
        # ax[1].set_xticklabels(df_apv['metric'].apply(lambda x: 'PCR #' + x), fontsize=7, rotation=30)
        ax[1].set_ylabel('UMI count', fontsize=14)
        ax[1].set_title('Merged' + ' (' + self.method + ')', fontsize=12)
        ax[1].spines['right'].set_visible(False)
        ax[1].spines['top'].set_visible(False)
        handles1, labels1 = ax[0].get_legend_handles_labels()
        ax[0].legend(
            handles1,
            labels1,
            fontsize=12,
        )
        handles2, labels2 = ax[1].get_legend_handles_labels()
        ax[1].legend(
            handles2,
            labels2,
            fontsize=12,
        )
        fig.subplots_adjust(
            top=0.95,
            bottom=0.1,
            left=0.18,
            right=0.95,
            hspace=0.20,
            # wspace=0.15
        )
        plt.show()

    def n2(self, df):
        fig, ax = plt.subplots(1, 1, figsize=(10, 6), sharex=True)
        df_gp = df.groupby(by=['method'])
        df_gp_keys = df_gp.groups.keys()
        method_map = {
            'ccs': 'cluster',
            'adj': 'adjacency',
            'direc': r'$directional$',
            'mcl_val': 'MCL-val',
            'mcl_ed': 'MCL-ed',
        }
        palette = {
            'ccs': 'red',
            'adj': 'black',
            'direc': 'steelblue',
            'mcl_val': 'chocolate',
            'mcl_ed': 'firebrick',
            # 'black',
            # 'chocolate',
            # 'saddlebrown',
            # 'darkgoldenrod',
            # 'firebrick',
        }
        for method in df_gp_keys:
            df_met = df_gp.get_group(method)
            print(df_met)
            # if method != 'adj':
            ax.plot(
                df_met['metric'],
                df_met['dedup_cnt'].apply(
                    # lambda x: x / 50
                    lambda x: (x - 50) / 50
                    # lambda x: np.exp((x - 50) / 50)
                ),
                label=method_map[method],
                color=palette[method],
                lw=2.5,
                alpha=0.7
            )
        # ax[0].set_xlabel('Time (ps)', fontsize=14)
        c = df.loc[df['method'] == 'mcl_ed']['metric']
        ax.set_xticks(c)
        ax.set_xticklabels(c, fontsize=8)
        # ax.set_xticklabels(c.astype(np.float).apply(lambda x: '{:.2e}'.format(x)), fontsize=8, rotation=30)
        # ax.set_xticklabels(c.astype(np.float).round(1), fontsize=8)

        # ax.set_xlabel('PCR cycle', fontsize=11)
        # ax.set_xlabel('Polymerase error', fontsize=11)
        # ax.set_xlabel('Sequencing error', fontsize=11)
        ax.set_xlabel('UMI length', fontsize=11)
        # ax.set_xlabel('Amplification rate', fontsize=11)
        ax.set_ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        # sns.lineplot(data=data, palette="tab10", linewidth=2.5)
        handles1, labels1 = ax.get_legend_handles_labels()
        ax.legend(
            handles1,
            labels1,
            fontsize=10,
        )
        fig.subplots_adjust(
            # top=0.92,
            # bottom=0.13,
            # left=0.13,
            # right=0.95,
            hspace=0.40,
            # wspace=0.15
        )
        plt.show()

    def n2dist(self, df):
        sns.displot(data=df, x='dedup_cnt', hue='method', kind="kde", rug=True)
        plt.show()


if __name__ == "__main__":
    from umiche.path import to

    from umiche.deduplicate.io.Stat import Stat as dedupstat

    scenarios = {
        'pcr_nums': 'PCR cycle',
        # 'pcr_errs': 'PCR error',
        # 'seq_errs': 'Sequencing error',
        # 'ampl_rates': 'Amplification rate',
        # 'umi_lens': 'UMI length',
        # 'seq_deps': 'Sequencing depth',
    }
    methods = {
        # 'unique': 'Unique',
        # 'cluster': 'Cluster',
        # 'adjacency': 'Adjacency',
        'directional': 'Directional',
        # 'dbscan_seq_onehot': 'DBSCAN',
        # 'birch_seq_onehot': 'Birch',
        # 'aprop_seq_onehot': 'Affinity Propagation',
        # 'mcl': 'MCL',
        # 'mcl_val': 'MCL-val',
        # 'mcl_ed': 'MCL-ed',
    }

    dedupstat = dedupstat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=to('data/params.yml'),
    )

    p = TraceSingle(
        df_apv=dedupstat.df_trace_cnt['apv'],
        df_disapv=dedupstat.df_trace_cnt['disapv'],
    )

    p.line_apv_disapv()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit


class split:

    def __init__(self,):
        pass

    def sss(self, data_dict, num_splits=5, train_ratio=1, test_ratio=1):
        """
        ..  Summary:
            --------
            split by the stratified shuffle split (SSS) method.

        ..  Description:
            ------------
            res is a dict
            {
                cv0: {
                    train: train_row_ids,
                    test: train_row_ids,
                    train_cls: train_cell_cls,
                    test_cls: test_cell_cls,
                },
                cv2: {
                    train: train_row_ids,
                    test: train_row_ids,
                    train_cls: train_cell_cls,
                    test_cls: test_cell_cls,
                }
                cv3: {
                    train: train_row_ids,
                    test: train_row_ids,
                    train_cls: train_cell_cls,
                    test_cls: test_cell_cls,
                },
                ...
                cvt: {
                    train: train_row_ids,
                    test: train_row_ids,
                    train_cls: train_cell_cls,
                    test_cls: test_cell_cls,
                }
            }

        :param data_dict:
        :param num_splits:
        :param train_ratio:
        :param test_ratio:
        :return:
        """
        df = pd.DataFrame.from_dict(data_dict['dict_cells'], orient='index', columns=['cluster'])
        uniq = df.cluster.unique()
        clusters = {e: i for i, e in enumerate(uniq)}
        clusters_ = {i: e for i, e in enumerate(uniq)}
        df['cls'] = df.cluster.apply(lambda x: clusters[x])
        # print(df)
        X = df.index.values
        y = df.cls.values
        sss = StratifiedShuffleSplit(
            n_splits=num_splits,
            test_size=test_ratio/(train_ratio + test_ratio),
            random_state=0,
        )
        res = {}
        for i_cv, (train_index, test_index) in enumerate(sss.split(X, y)):
            print("Train:", train_index.shape, "Test:", test_index.shape)
            df_cv_train = df.iloc[train_index]
            df_cv_test = df.iloc[test_index]
            df_cv_train_gp = df_cv_train.groupby(by=['cls'])
            df_cv_test_gp = df_cv_test.groupby(by=['cls'])
            cv_train_keys = df_cv_train_gp.groups.keys()
            cv_test_keys = df_cv_test_gp.groups.keys()
            # print(cv_train_keys)
            # print(cv_test_keys)
            cv_train_cls_dict = {clusters_[i]: df_cv_train_gp.get_group(i).index.values.astype(int) for i in cv_train_keys}
            cv_test_cls_dict = {clusters_[i]: df_cv_test_gp.get_group(i).index.values.astype(int) for i in cv_test_keys}
            res['cv' + str(i_cv)] = {
                'train_index': train_index,
                'test_index': test_index,
                'train_cls': cv_train_cls_dict,
                'test_cls': cv_test_cls_dict,
                'train_cls_all': df.iloc[train_index].cls.values,
                'test_cls_all': df.iloc[test_index].cls.values,
            }
        # print(res)
        return res

    def kfold(self, res):
        pass


if __name__ == "__main__":
    p = split()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
from umiche.util.Console import Console


class Edge:

    def __init__(
            self,
            graph=None,
            verbose=False,
    ):
        self._graph = graph
        self._graph_mapped = None

        self.console = Console()
        self.console.verbose = verbose

    @property
    def graph(self, ):
        self.console.print('============>the current graph is {}'.format(self._graph))
        return self._graph

    @property
    def glen(self, ):
        self.console.print('============>the number of egdes in the current graph is {}'.format(len(self._graph)))
        return len(self._graph)

    @property
    def nodes(self, ):
        g_np = np.array(self._graph)
        l = np.array(g_np)[:, 0]
        r = np.array(g_np)[:, 1]
        return list(np.unique(np.concatenate((l, r))))

    @property
    def key_mapped(self, ):
        return {k: id for id, k in enumerate(self.nodes)}

    @graph.setter
    def graph(self, value):
        self._graph = value

    @property
    def graph_mapped(self, ):
        if self._graph == None:
            raise 'go set your graph'
        else:
            return self.map(self._graph)

    @property
    def rvredanduncy(self, ):
        repeat = []
        edges = self._graph.copy()
        edge_set = set(self._graph)
        while edges:
            edge = edges.pop(0)
            if tuple(reversed(edge)) in edges:
                repeat.append(edge)
        edges = list(edge_set.difference(set(repeat)))
        # print(self._graph)
        return edges

    def map(self, graph):
        self.console.print('============>>the graph is being mapped')
        self.console.print('============>>key map:\n {}'.format(self.key_mapped))
        g_mapped = []
        for i, edge in enumerate(graph):
            g_mapped.append((self.key_mapped[edge[0]], self.key_mapped[edge[1]]))
        self.console.print('============>>the mapped graph is\n {}'.format(g_mapped))
        return g_mapped

    def toAdjacencyDict(self, ):
        adj_list = {}
        # scan the arrays edge_u and edge_v
        for i in self.nodes:
            adj_list[i] = []
        for i in range(self.glen):
            l = self._graph[i][0]
            r = self._graph[i][1]
            adj_list[l].append(r)
            adj_list[r].append(l)
        return adj_list

    def fromlist(self, list_2d):
        tuples = []
        for e in list_2d:
            tuples.append((e[0], e[1]))
        return tuples

    def fromdf(self, df, col_name1='node_1', col_name2='node_2', to_tuple=False):
        edge_list = df[[col_name1, col_name2]].values.tolist()
        if to_tuple:
            return self.fromlist(edge_list)
        else:
            return edge_list

    def toEdgeList(
            self,
            graph_adj,
            rr=True,
    ):
        edges = []
        for k, vals in graph_adj.items():
            for val in vals:
                edges.append((k, val))
        if rr:
            repeat = []
            edge_set = set(edges)
            while edges:
                edge = edges.pop(0)
                if tuple(reversed(edge)) in edges:
                    repeat.append(edge)
            edges = list(edge_set.difference(set(repeat)))
        return edges


if __name__ == "__main__":
    graph_edge_list = [
        ('A', 'B'),
        ('A', 'C'),
        ('B', 'C'),
        ('A', 'D'),
        ('D', 'E'),
        ('D', 'F'),
        ('D', 'F'),
    ]

    p = Edge(
        # graph=graph_edge_list,
    )

    p.graph = graph_edge_list

    print(p.graph)

    print(p.graph_mapped)

    p.graph = p.graph_mapped

    print(p.graph)

    print(p.key_mapped)

    p.graph = p.rvredanduncy

    print(p.graph)

    # print(p.key_mapped)

    # print(p.rvredanduncy)

    print('asd', p.toAdjacencyDict())

    # print(p.graph_mapped)

    # print(p.nodes)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
from umiche.util.Console import Console


class Adjacency:

    def __init__(
            self,
            graph=None,
            verbose=True,
    ):
        self._graph = graph
        self._graph_mapped = None

        self.console = Console()
        self.console.verbose = verbose

    @property
    def graph(self, ):
        self.console.print('===>The current graph: {}'.format(self._graph))
        return self._graph

    @graph.setter
    def graph(
            self,
            value,
    ):
        """

        Parameters
        ----------
        value

        Returns
        -------

        """
        self._graph = value

    @property
    def graph_mapped(self, ):
        if self._graph == None:
            raise 'Please set and input your graph'
        else:
            return self.map(self._graph)

    @property
    def glen(self, ):
        # print('the number of egdes in the current graph is {}'.format(len(self._graph)))
        return len(self._graph)

    @property
    def key_mapped(self, ):
        return {[*self._graph.keys()][k]: k for k in range(len([*self._graph.keys()]))}

    def map(
            self,
            graph,
    ):
        """

        Parameters
        ----------
        graph

        Returns
        -------

        """
        self.console.print('===>the graph is being mapped')
        self.console.print('======>key map: {}'.format(self.key_mapped))
        if isinstance(graph, dict):
            self.console.print('======>the graph is a dict')
            g_mapped = {}
            for k, vals in self._graph.items():
                g_mapped[self.key_mapped[k]] = []
                for val in vals:
                    g_mapped[self.key_mapped[k]].append(self.key_mapped[val])
            self.console.print('======>the mapped graph: {}'.format(g_mapped))
            return g_mapped

    def dict(self, ):
        return self._graph

    def set(self, ):
        adj_set = {}
        for k, vals in self._graph.items():
            adj_set[k] = set(vals)
        return adj_set

    def list(self, ):
        return [*self._graph.values()]

    def to_matrix(self, ):
        adj_mat = np.zeros(shape=[self.glen, self.glen])
        for k, vals in self._graph.items():
            for val in vals:
                adj_mat[self.key_mapped[k], self.key_mapped[val]] = 1
        return adj_mat

    def hash(self, ):
        return

    def to_edge_list(
            self,
            rr=True,
    ):
        """

        Parameters
        ----------
        rr
            if it is not symmetry.
            False
            [(0, 1), (0, 2), (0, 3), (1, 0), (1, 2), (2, 0), (2, 1), (3, 0), (3, 4), (3, 5), (4, 3), (4, 6), (5, 3), (5, 6), (6, 4), (6, 5)]
            True
            [(2, 1), (6, 5), (4, 3), (2, 0), (6, 4), (3, 0), (1, 0), (5, 3)]

        Returns
        -------

        """
        edges = []
        for k, vals in self._graph.items():
            for val in vals:
                edges.append((k, val))
        if rr:
            repeat = []
            edge_set = set(edges)
            while edges:
                edge = edges.pop(0)
                if tuple(reversed(edge)) in edges:
                    repeat.append(edge)
            edges = list(edge_set.difference(set(repeat)))
        return edges


if __name__ == "__main__":
    # ### @@ data from UMI-tools
    # graph_adj_dict = {
    #     'A': ['B', 'C'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['E', 'F'],
    #     'E': ['D'],
    #     'F': ['D'],
    # }

    ### @@ data from mine
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }

    p = Adjacency(graph_adj)

    p.graph = p.graph_mapped

    print(p.graph)

    print(p.dict())

    print(p.set())

    print(p.list())

    print(p.to_matrix())

    print(p.to_edge_list(rr=False))
    # print(p.to_edge_list(rr=True))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import List, Dict

import pandas as pd

from umiche.deduplicate.method.Cluster import Cluster
from umiche.deduplicate.method.Adjacency import Adjacency
from umiche.deduplicate.method.Directional import Directional
from umiche.deduplicate.method.MarkovClustering import MarkovClustering
from umiche.deduplicate.method.Clustering import Clustering
from umiche.deduplicate.method.trimer.MajorityVote import MajorityVote
from umiche.deduplicate.method.trimer.SetCover import SetCover


def cluster(
        graph,
        method : str = 'deque',
        verbose : bool = False,
):
    if method == 'deque':
        return Cluster(
            verbose=verbose,
        ).cc(
            graph_adj=graph,
        )
    if method == 'networkx':
        return Cluster(
            verbose=verbose,
        ).ccnx(
            edge_list=graph,
        )


def adjacency(
        connected_components : Dict,
        df_umi_uniq_val_cnt : pd.Series,
        graph_adj : Dict,
        verbose: bool = False,
):
    return Adjacency(
        verbose=verbose,
    ).umi_tools(
        connected_components=connected_components,
        df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
        graph_adj=graph_adj,
    )


def directional(
        connected_components : Dict,
        df_umi_uniq_val_cnt : pd.Series,
        graph_adj : Dict,
        verbose: bool = False,
):
    return Directional(
        verbose=verbose,
    ).umi_tools(
        connected_components=connected_components,
        df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
        graph_adj=graph_adj,
    )


def mcl(
        inflat_val : float,
        exp_val : int,
        iter_num : int,
        connected_components : Dict,
        graph_adj : Dict,
        verbose: bool = False,
):
    return MarkovClustering(
        inflat_val=inflat_val,
        exp_val=exp_val,
        iter_num=iter_num,
        verbose=verbose,
    ).dfclusters(
        connected_components=connected_components,
        graph_adj=graph_adj,
    )


def mcl_val(
        df_mcl_ccs : pd.DataFrame,
        df_umi_uniq_val_cnt : pd.Series,
        thres_fold : int,
        verbose: bool = False,
):
    return MarkovClustering(
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
        verbose=verbose,
    ).maxval_val(
        df_mcl_ccs=df_mcl_ccs,
        df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
        thres_fold=thres_fold,
    )


def mcl_ed(
        df_mcl_ccs : pd.DataFrame,
        df_umi_uniq_val_cnt : pd.Series,
        thres_fold : int,
        int_to_umi_dict : Dict,
        verbose: bool = False,
):
    return MarkovClustering(
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
        verbose=verbose,
    ).maxval_ed(
        df_mcl_ccs=df_mcl_ccs,
        df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
        thres_fold=thres_fold,
        int_to_umi_dict=int_to_umi_dict,
    )


def dbscan(
        connected_components : Dict,
        graph_adj : Dict,
        int_to_umi_dict : Dict,
        dbscan_eps=1.5,
        dbscan_min_spl=1,
        verbose: bool = False,
):
    return Clustering(
        clustering_method='dbscan',
        dbscan_eps=dbscan_eps,
        dbscan_min_spl=dbscan_min_spl,
        verbose=verbose,
    ).dfclusters(
        connected_components=connected_components,
        graph_adj=graph_adj,
        int_to_umi_dict=int_to_umi_dict,
    )


def birch(
        connected_components : Dict,
        graph_adj : Dict,
        int_to_umi_dict : Dict,
        birch_thres=1.8,
        birch_n_clusters=None,
        verbose: bool = False,
):
    return Clustering(
        clustering_method='birch',
        birch_thres=birch_thres,
        birch_n_clusters=birch_n_clusters,
        verbose=verbose,
    ).dfclusters(
        connected_components=connected_components,
        graph_adj=graph_adj,
        int_to_umi_dict=int_to_umi_dict,
    )


def affinity_propagation(
        connected_components : Dict,
        graph_adj : Dict,
        int_to_umi_dict : Dict,
        verbose: bool = False,
):
    return Clustering(
        clustering_method='aprop',
        verbose=verbose,
    ).dfclusters(
        connected_components=connected_components,
        graph_adj=graph_adj,
        int_to_umi_dict=int_to_umi_dict,
    )


def set_cover(
        multimer_list,
        recur_len=3,
        split_method='split_to_all',
        verbose=True,
):
    return SetCover(
        verbose=verbose,
    ).greedy(
        multimer_list=multimer_list,
        recur_len=recur_len,
        split_method=split_method,
    )


def majority_vote(
        multimer_list,
        recur_len=3,
        verbose=True,
):
    return MajorityVote(
        verbose=verbose,
    ).track(
        multimer_list=multimer_list,
        recur_len=recur_len,
    )


def decompose(
        cc_sub_dict : Dict,
        verbose: bool = False,
):
    return Adjacency(
        verbose=verbose,
    ).decompose(
        cc_sub_dict=cc_sub_dict,
    )


def decompose_mcl(
        list_nd : List,
        verbose: bool = False,
):
    return MarkovClustering(
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
        verbose=verbose,
    ).decompose(
        list_nd=list_nd,
    )


if __name__ == "__main__":
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }

    ccs = cluster(
        graph=graph_adj,
        method='deque',
    )
    # print(ccs)

    # edge_list = [('C', 'B'), ('C', 'A'), ('B', 'A'), ('G', 'E'), ('E', 'D'), ('G', 'F'), ('D', 'A'), ('F', 'D')]
    #
    # print(cluster(
    #     graph=edge_list,
    #     method='networkx',
    # ))

    node_val_sorted = pd.Series({
        'A': 120,
        'D': 90,
        'E': 50,
        'G': 5,
        'B': 2,
        'C': 2,
        'F': 1,
    })

    ### ++++++++++++++adjacency++++++++++++++++++++++
    # dedup_res = adjacency(
    #     connected_components=ccs,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     graph_adj=graph_adj,
    #     verbose=True,
    # )
    # dedup_count = dedup_res['count']
    # dedup_clusters = dedup_res['clusters']
    # print("deduplicated count:\n{}".format(dedup_count))
    # print("deduplicated clusters:\n{}".format(dedup_clusters))
    # dedup_clusters_dc = decompose(dedup_clusters)
    # print("deduplicated clusters decomposed:\n{}".format(dedup_clusters_dc))

    ### ++++++++++++++directional++++++++++++++++++++++
    # dedup_res = directional(
    #     connected_components=ccs,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     graph_adj=graph_adj,
    #     verbose=True,
    # )
    # dedup_count = dedup_res['count']
    # dedup_clusters = dedup_res['clusters']
    # print("deduplicated count:\n{}".format(dedup_count))
    # print("deduplicated clusters:\n{}".format(dedup_clusters))
    # dedup_clusters_dc = decompose(dedup_clusters)
    # print("deduplicated clusters decomposed:\n{}".format(dedup_clusters_dc))
    #
    # print(dedup_res['apv'])
    # print(dedup_res['disapv'])

    # ### ++++++++++++++MCL++++++++++++++++++++++
    # dedup_res = mcl(
    #     inflat_val=1.6,
    #     exp_val=2,
    #     iter_num=100,
    #     connected_components=ccs,
    #     graph_adj=graph_adj,
    #     verbose=True,
    # )
    # print(dedup_res.columns)
    # print("vertices: {}".format(dedup_res.loc[0, 'cc_vertices']))
    # print(dedup_res.loc[0, 'graph_cc_adj'])
    # print(dedup_res.loc[0, 'nt_to_int_map'])
    # print(dedup_res.loc[0, 'int_to_nt_map'])
    # print(dedup_res.loc[0, 'cc_adj_mat'])
    # print(dedup_res.loc[0, 'mcl_clusters'])
    # print(dedup_res.loc[0, 'clusters'])
    # print(dedup_res.loc[0, 'clust_num'])
    #
    # print(dedup_res['clusters'].values)
    # dedup_clusters_dc = decompose_mcl(list_nd=dedup_res['clusters'].values)
    # print("deduplicated clusters decomposed (mcl):\n{}".format(dedup_clusters_dc))

    ### ++++++++++++++MCL-val++++++++++++++++++++++
    # df_mcl_val = mcl_val(
    #     df_mcl_ccs=dedup_res,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     thres_fold=1,
    # )
    # print(df_mcl_val)
    # dedup_count = df_mcl_val['count'].values[0]
    # dedup_clusters = df_mcl_val['clusters'].values[0]
    # print("deduplicated count (mcl_val):\n{}".format(dedup_count))
    # print("deduplicated clusters (mcl_val):\n{}".format(dedup_clusters))
    #
    # df_mcl_val = decompose_mcl(list_nd=df_mcl_val['clusters'].values)
    # print("deduplicated clusters decomposed (mcl_val):\n{}".format(df_mcl_val))

    ### ++++++++++++++MCL-ed++++++++++++++++++++++
    # int_to_umi_dict = {
    #     'A': 'AGATCTCGCA',
    #     'B': 'AGATCCCGCA',
    #     'C': 'AGATCACGCA',
    #     'D': 'AGATCGCGCA',
    #     'E': 'AGATCGCGGA',
    #     'F': 'AGATCGCGTA',
    #     'G': 'TGATCGCGAA',
    # }
    #
    # df_mcl_ed = mcl_ed(
    #     df_mcl_ccs=dedup_res,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     thres_fold=1,
    #     int_to_umi_dict=int_to_umi_dict,
    # )
    # dedup_count = df_mcl_ed['count']
    # dedup_clusters = df_mcl_ed['clusters']
    # print('approval: {}'.format(df_mcl_ed['apv']))
    #
    # print("deduplicated count (mcl_ed):\n{}".format(dedup_count))
    # print("deduplicated clusters (mcl_ed):\n{}".format(dedup_clusters))
    #
    # df_mcl_ed = decompose_mcl(list_nd=df_mcl_ed['clusters'].values)
    # print("deduplicated clusters decomposed (mcl_ed):\n{}".format(df_mcl_ed))

    # ### ++++++++++++++dbscan++++++++++++++++++++++
    # int_to_umi_dict = {
    #     'A': 'AGATCTCGCA',
    #     'B': 'AGATCCCGCA',
    #     'C': 'AGATCACGCA',
    #     'D': 'AGATCGCGCA',
    #     'E': 'AGATCGCGGA',
    #     'F': 'AGATCGCGTA',
    #     'G': 'TGATCGCGAA',
    # }
    # df = dbscan(
    #     connected_components=ccs,
    #     graph_adj=graph_adj,
    #     # df_umi_uniq_val_cnt=node_val_sorted,
    #     int_to_umi_dict=int_to_umi_dict,
    #     dbscan_eps=1.5,
    #     dbscan_min_spl=1,
    # )
    # # print(df)
    # for i, col in enumerate(df.columns):
    #     print("{}.{}: \n{}".format(i+1, col, df[col].values[0]))
    # df_decomposed = decompose_mcl(list_nd=df['clusters'].values)
    # print("deduplicated clusters decomposed:\n{}".format(df_decomposed))

    # ### ++++++++++++++birch++++++++++++++++++++++
    # int_to_umi_dict = {
    #     'A': 'AGATCTCGCA',
    #     'B': 'AGATCCCGCA',
    #     'C': 'AGATCACGCA',
    #     'D': 'AGATCGCGCA',
    #     'E': 'AGATCGCGGA',
    #     'F': 'AGATCGCGTA',
    #     'G': 'TGATCGCGAA',
    # }
    # df = birch(
    #     connected_components=ccs,
    #     graph_adj=graph_adj,
    #     int_to_umi_dict=int_to_umi_dict,
    #     birch_thres=1.8,
    #     birch_n_clusters=None,
    # )
    # # print(df)
    # for i, col in enumerate(df.columns):
    #     print("{}.{}: \n{}".format(i + 1, col, df[col].values[0]))
    # df_decomposed = decompose_mcl(list_nd=df['clusters'].values)
    # print("deduplicated clusters decomposed:\n{}".format(df_decomposed))

    ### ++++++++++++++birch++++++++++++++++++++++
    # int_to_umi_dict = {
    #     'A': 'AGATCTCGCA',
    #     'B': 'AGATCCCGCA',
    #     'C': 'AGATCACGCA',
    #     'D': 'AGATCGCGCA',
    #     'E': 'AGATCGCGGA',
    #     'F': 'AGATCGCGTA',
    #     'G': 'TGATCGCGAA',
    # }
    # df = affinity_propagation(
    #     connected_components=ccs,
    #     graph_adj=graph_adj,
    #     int_to_umi_dict=int_to_umi_dict,
    # )
    # # print(df)
    # for i, col in enumerate(df.columns):
    #     print("{}.{}: \n{}".format(i + 1, col, df[col].values[0]))
    # df_decomposed = decompose_mcl(list_nd=df['clusters'].values)
    # print("deduplicated clusters decomposed:\n{}".format(df_decomposed))

    # ### ++++++++++++++set_cover++++++++++++++++++++++
    # from umiche.bam.Reader import Reader as alireader
    # alireader = alireader(
    #     bam_fpn="/mnt/d/Document/Programming/Python/umiche/umiche/data/simu/umi/trimer/seq_errs/permute_0/trimmed/seq_err_17.bam",
    #     verbose=True,
    # )
    # df_bam = alireader.todf(tags=['PO'])
    # print(df_bam.columns)
    # trimer_list = df_bam.query_name.apply(lambda x: x.split('_')[1]).values
    # print(trimer_list)
    # (
    #     dedup_cnt,
    #     multimer_umi_solved_by_sc,
    #     multimer_umi_not_solved,
    #     shortlisted_multimer_umi_list,
    #     monomer_umi_lens,
    #     multimer_umi_lens,
    # ) = set_cover(
    #     multimer_list=trimer_list,
    #     recur_len=3,
    #     split_method='split_to_all',
    #     verbose=True
    # )
    # print(dedup_cnt)

    ### ++++++++++++++set_cover++++++++++++++++++++++
    from umiche.bam.Reader import Reader as alireader
    alireader = alireader(
        bam_fpn="/mnt/d/Document/Programming/Python/umiche/umiche/data/simu/umi/trimer/seq_errs/permute_0/trimmed/seq_err_17.bam",
        verbose=True,
    )
    df_bam = alireader.todf(tags=['PO'])
    print(df_bam.columns)
    trimer_list = df_bam.query_name.apply(lambda x: x.split('_')[1]).values
    print(trimer_list)
    (
        dedup_cnt,
        uniq_multimer_cnt,
        shortlisted_multimer_umi_list,
    ) = majority_vote(
        multimer_list=trimer_list,
        recur_len=3,
        verbose=True
    )
    print(dedup_cnt)
    print(uniq_multimer_cnt)
    print(shortlisted_multimer_umi_list)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pyfastx


class Reader():

    def __init__(self):
        pass

    def fromgz(self, fastq_fpn):
        """
        read and parse a Fastq file with pyfastx.

        Parameters
        ----------
        fastq_fpn
            the path of a fastq file
        Returns
        -------
            tuple consisting of names, seqs, placeholders, qualities
        """
        names = []
        seqs = []
        placeholders = []
        qualities = []
        fq = pyfastx.Fastx(fastq_fpn)
        # print(seqs)
        for name, seq, qual in fq:
            seqs.append(''.join(seq))
            names.append(''.join(name))
        return names, seqs, placeholders, qualities



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import to_categorical
from umiche.deepconvcvae.Preprocess import preprocess as datapp
from umiche.deepconvcvae.CondiConvVAE import condiConvVAE
from umiche.plot.scatter.DimensionReduction import dimensionReduction as drplot


def sampling(
        decoder,
        num_labels,
        image_size,
        batch_size=32,
):
    xs = []
    ys = []
    # num_ = 100
    num_ = 1000
    for ilabel in range(num_labels):
        print('===>label {}'.format(ilabel))
        digit_size = image_size
        z_combo_spl = np.random.normal(loc=1, scale=4, size=[num_, 2])
        # z_combo_spl = np.random.uniform(0, 4, size=[num_, 2])
        ilabel_ = np.eye(num_labels)[np.array([ilabel])]
        for i, zcspl in enumerate(z_combo_spl):
            z_sample = np.array([zcspl])
            x_decoded = decoder.predict([z_sample, ilabel_])
            # print()
            digit = x_decoded[0].reshape(digit_size, digit_size)
            xs.append(digit.reshape(digit_size * digit_size))
            # xs.append(x_decoded[0])
            ys.append(ilabel)
            # sds = digit * 255
            # print(np.where(sds > 1, sds, 0))
    drplot().single(X=xs, y=ys, tech='TSNE', marker_size=3, cmap='tab20b', title='CondiCVAE')
    drplot().single(X=xs, y=ys, tech='PCA', marker_size=3, cmap='tab20b', title='CondiCVAE')
    drplot().single(X=xs, y=ys, tech='UMAP', marker_size=3, cmap='tab20b', title='CondiCVAE')


def data_imputation(
        image_size,
        params,
):
    pp = datapp(params['data_rv_zero_fpn'], train_ratio=3, test_ratio=2)
    (x_train, y_train), (x_test, y_test) = pp.recipeData(cv=0)
    # print(x_train.shape)
    # print(y_train.shape)
    x_train_height = x_train.shape[0]
    x_test_height = x_test.shape[0]
    x_train_width = x_train.shape[1]

    img_col_offset_size = image_size*image_size - x_train_width
    train_col_offset_mat = np.zeros([x_train_height, img_col_offset_size])
    test_col_offset_mat = np.zeros([x_test_height, img_col_offset_size])
    x_train = np.concatenate([x_train, train_col_offset_mat], axis=1)
    x_test = np.concatenate([x_test, test_col_offset_mat], axis=1)
    x_train = np.reshape(x_train, newshape=(-1, image_size, image_size))
    x_test = np.reshape(x_test, newshape=(-1, image_size, image_size))
    # print(x_train)
    # print(x_test)
    x_train = np.reshape(x_train, [-1, image_size, image_size, 1])
    x_test = np.reshape(x_test, [-1, image_size, image_size, 1])
    print(x_train.shape)
    print(y_train.shape)
    print(x_test.shape)
    print(y_test.shape)
    print(np.amax(x_train))
    print(np.amax(x_test))
    std_max_mark = np.maximum(np.amax(x_train), np.amax(x_test))
    print(std_max_mark)
    x_train = x_train.astype('float32') / std_max_mark
    x_test = x_test.astype('float32') / std_max_mark
    # print(y_train)
    # print(y_test)
    return x_train, y_train, x_test, y_test


def train(
        x_train,
        y_train,
        x_test,
        y_test,
        image_size,
        sv_fpn,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        strides=2,
        epochs=1,
):
    # num_labels = 10
    num_labels = len(np.unique(y_train))
    # print(num_labels)
    input_shape = (image_size, image_size, 1)
    label_shape = (num_labels,)

    inputs = tf.keras.layers.Input(shape=input_shape, name='encoder_input')
    y_labels = tf.keras.layers.Input(shape=label_shape, name='class_labels')
    print(inputs)
    print(y_labels)
    condicvae = condiConvVAE(
        image_size=image_size,
        input_shape=input_shape,
        label_shape=label_shape,
        batch_size=batch_size,
        kernel_size=kernel_size,
        filters=filters,
        latent_dim=latent_dim,
        strides=strides,
        epochs=epochs,
        inputs=inputs,
        y_labels=y_labels,
    )
    shape, z_mean, z_log_var, encoder = condicvae.encoding()
    # encoder.summary()
    decoder = condicvae.decoding(shape)
    # decoder.summary()

    outputs = decoder([encoder([inputs, y_labels])[2], y_labels])
    ccvae = tf.keras.models.Model(inputs=[inputs, y_labels], outputs=outputs, name='ccvae')
    ccvae.add_loss(condicvae.loss(outputs, z_mean, z_log_var))
    ccvae.compile(optimizer='rmsprop')
    ccvae.summary()

    ccvae.fit(
        [x_train, to_categorical(y_train)],
        epochs=epochs,
        batch_size=batch_size,
        validation_data=([x_test, to_categorical(y_test)], None),
    )

    # model_name = "cvae_model20"
    # save_dir = "cvae_model20"
    # if not os.path.isdir(save_dir):
    #     os.makedirs(save_dir)
    # filename = model_name + '.tf'
    # filepath = os.path.join(save_dir, filename)
    # ccvae.save_weights(filepath)

    ccvae.save(sv_fpn)

    sampling(
        decoder,
        num_labels=num_labels,
        batch_size=batch_size,
        image_size=image_size,
    )
    return


if __name__ == "__main__":
    from umiche.path import to

    params = {
        'data_fpn': to('data/deepconvcvae/flt.h5'),
        'data_rv_zero_fpn': to('data/deepconvcvae/flt_rv_zero_col.h5'),
    }

    image_size = 144

    x_train, y_train, x_test, y_test = data_imputation(
        params=params,
        image_size=image_size,
    )

    print(train(
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        image_size=image_size,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        strides=2,
        epochs=1,
        sv_fpn='./cvae',
    ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import pandas as pd
from umiche.fastq.Reader import Reader as rfastq
from umiche.fastq.Writer import Writer as wfastq

from umiche.trim.Reader import Reader as trimreader
from umiche.simu.Parameter import Parameter as params

from umiche.util.Writer import Writer as fwriter
from umiche.util.Console import Console


class Shakeup:

    def __init__(
            self,
            scenario,
            umi_len=36,

            param_fpn=None,
            is_trim=False,
            is_split=False,

            verbose=False,

            **kwargs,
    ):
        self.scenario = scenario
        self.umi_len = umi_len
        self.is_trim = is_trim
        self.is_split = is_split
        self.kwargs = kwargs

        self.params = params(param_fpn=param_fpn)

        self.rfastq = rfastq()
        self.wfastq = wfastq()
        self.trimreader = trimreader()
        self.fwriter = fwriter()

        self.verbose = verbose
        self.console = Console()
        self.console.verbose = self.verbose


        for perm_num_i in range(self.params.fixed['permutation_num']):
            for id, scenario_i in enumerate(self.params.varied[self.scenario]):
                read_stime = time.time()
                print('=========>permutation {} under scenario No.{} {}'.format(perm_num_i, id, scenario_i))

                if self.scenario == 'pcr_nums':
                    self.fn_mark = str(id) # scenario_i
                elif self.scenario == 'pcr_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'ampl_rates':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_deps':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_lens':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_nums':
                    self.fn_mark = str(id)
                else:
                    self.fn_mark = str(scenario_i)
                self.fn_prefix = self.params.file_names[self.scenario] + self.fn_mark
                self.fastq_location = self.params.work_dir + self.scenario + '/permute_' + str(perm_num_i) + '/'

                print(self.fn_prefix)
                print(self.fastq_location)

                if self.is_trim:
                    from umiche.trim.Template import Template as trimmer
                    if self.scenario == 'umi_lens':
                        self.params.trimmed['umi_1']['len'] = scenario_i
                    self.console.print("======>fastq is being trimmed.")
                    self.params.trimmed['fastq']['fpn'] = self.fastq_location + self.fn_prefix + '.fastq.gz'
                    self.params.trimmed['fastq']['trimmed_fpn'] = self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz'
                    umitrim_parser = trimmer(params=self.params.trimmed, verbose=self.verbose)
                    df = umitrim_parser.todf()
                    umitrim_parser.togz(df)
                if self.is_split:
                    names, seqs, _, _ = self.rfastq.fromgz(
                        fastq_fpn=self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz'
                    )
                    # print(seqs)
                    # print(names)

                    # df = self.trimreader.todf(names=names, seqs=seqs)
                    df = self.trimreader.todfFromTree(names=names, seqs=seqs)

                    # df['origin_info'] = df['name'].apply(lambda x: x.split('_')[0])
                    df['origin_info'] = df.apply(lambda x: str(x['umi#']) + '_' + x['umi_src'], axis=1)
                    mono_corr_stime = time.time()

                    df['umi_order'] = df['umi'].apply(lambda x: self.split_by_order(x, umi_len=self.umi_len))
                    df['umi_l_order'] = df['umi_order'].apply(lambda x: x.split(';')[0])
                    df['umi_m_order'] = df['umi_order'].apply(lambda x: x.split(';')[1])
                    df['umi_r_order'] = df['umi_order'].apply(lambda x: x.split(';')[2])
                    # print(df['umi_r_order'])
                    df['to_fas_l_order'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_l_order'], axis=1)
                    df['to_fas_m_order'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_m_order'], axis=1)
                    df['to_fas_r_order'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_r_order'], axis=1)
                    df_lmr = pd.concat([
                        pd.DataFrame({'seq_raw': df['seq_raw'].values, 'to_fas': df['to_fas_l_order'].values}),
                        pd.DataFrame({'seq_raw': df['seq_raw'].values, 'to_fas': df['to_fas_m_order'].values}),
                        pd.DataFrame({'seq_raw': df['seq_raw'].values, 'to_fas': df['to_fas_r_order'].values}),
                    ], axis=0).reset_index(drop=True)
                    # print(df_lmr)

                    df['umi_mv'] = df['umi'].apply(lambda x: self.split_by_mv(x, umi_len=self.umi_len))
                    df['umi_ref'] = df['umi_mv'].apply(lambda x: x.split(';')[0])
                    df['umi_l_mv'] = df['umi_mv'].apply(lambda x: x.split(';')[1])
                    df['umi_m_mv'] = df['umi_mv'].apply(lambda x: x.split(';')[2])
                    df['umi_r_mv'] = df['umi_mv'].apply(lambda x: x.split(';')[3])

                    df['to_fas'] = df.apply(lambda x: x['origin_info'] + '_' + x['umi_ref'], axis=1)
                    # print(df['to_fas'])
                    df['umi_mark'] = df['umi'].apply(lambda x: self.marker(x, umi_len=self.umi_len))
                    df_differ_3nt = df.loc[df['umi_mark'] == 1]


                    df_3differ_cp = df_differ_3bases.copy()

                    df_3differ = df_3differ.drop('umi_r_mv', 1)
                    df_3differ_cp = df_3differ_cp.drop('umi_m_mv', 1)
                    df_3differ = df_3differ.rename(columns={"umi_m_mv": "umi_bi"})
                    df_3differ_cp = df_3differ_cp.rename(columns={"umi_r_mv": "umi_bi"})

                    df_umi_3differ = pd.concat(
                        [df_3differ, df_3differ_cp],
                        axis=0,
                    ).reset_index(drop=True)
                    # print(df_umi_3differ)

                    if not df_umi_3differ.empty:
                        df_umi_3differ['to_fas'] = df_umi_3differ.apply(lambda x: x['origin_info'] + '_' + x['umi_bi'], axis=1)
                        print(df_umi_3differ)
                    # print(df_umi_3differ['umi_bi'])

                    df_not3differ = df.loc[df['umi_mark'] != 1]
                    df_not3differ['to_fas'] = df_not3differ.apply(lambda x: x['origin_info'] + '_' + x['umi_l_mv'], axis=1)
                    if df_umi_3differ.empty:
                        df_merge = df_not3differ[['seq_raw', 'to_fas']].reset_index(drop=True)
                    else:
                        df_merge = pd.concat(
                            [df_umi_3differ[['seq_raw', 'to_fas']], df_not3differ[['seq_raw', 'to_fas']]],
                            axis=0,
                        ).reset_index(drop=True)
                    print(df_merge)
                    self.wfastq.togz(
                        list_2d=df_lmr[['seq_raw', 'to_fas']].values,
                        sv_fpn=self.fastq_fp + 'seq_errs/permute_' + str(perm_num_i) + '/lmr/',
                        fn=self.cat + '_' + str(id),
                    )
                    self.wfastq.togz(
                        list_2d=df[['seq_raw', 'to_fas']].values,
                        sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(perm_num_i) + '/ref/',
                        fn=self.cat + '_' + str(id),
                    )
                    self.wfastq.togz(
                        list_2d=df_merge.values,
                        sv_fp=self.fastq_fp + 'seq_errs/permute_' + str(perm_num_i) + '/bipartite/',
                        fn=self.cat + '_' + str(id),
                    )
                    print('===>getting it done with time: {:.3f}s'.format(time.time() - mono_corr_stime))

    def split_by_mv(
            self,
            umi,
            umi_len=36,
    ):
        """

        Parameters
        ----------
        umi

        umi_len

        Returns
        -------

        """
        vernier = [i for i in range(umi_len) if i % 3 == 0]
        umi_trimers = [umi[v: v+3] for v in vernier]
        ref = []
        l = []
        m = []
        r = []
        for umi_trimer in umi_trimers:
            ref.append(umi_trimer[0])
            s = set(umi_trimer)
            if len(s) == 3:
                l.append(umi_trimer[0])
                m.append(umi_trimer[1])
                r.append(umi_trimer[2])
            elif len(s) == 2:
                sdict = {umi_trimer.count(i): i for i in s}
                l.append(sdict[2])
                m.append(sdict[2])
                r.append(sdict[2])
            else:
                l.append(umi_trimer[0])
                m.append(umi_trimer[0])
                r.append(umi_trimer[0])
        ref = ''.join(ref)
        l = ''.join(l)
        m = ''.join(m)
        r = ''.join(r)
        return ref + ';' + l + ';' + m + ';' + r

    def split_by_order(
            self,
            umi,
            umi_len=36,
    ):
        vernier = [i for i in range(umi_len) if i % 3 == 0]
        umi_trimers = [umi[v: v + 3] for v in vernier]
        l = []
        m = []
        r = []
        for umi_trimer in umi_trimers:
            l.append(umi_trimer[0])
            m.append(umi_trimer[1])
            r.append(umi_trimer[2])
        l = ''.join(l)
        m = ''.join(m)
        r = ''.join(r)
        return l + ';' + m + ';' + r

    def marker(
            self,
            umi,
            umi_len=36,
    ):
        vernier = [i for i in range(umi_len) if i % 3 == 0]
        umi_trimers = [umi[v: v+3] for v in vernier]
        slens = [len(set(umi_trimer)) for umi_trimer in umi_trimers]
        if 3 in slens:
            return 1
        else:
            return 0


if __name__ == "__main__":
    from umiche.path import to

    p = Shakeup(
        # scenario='pcr_nums',
        # scenario='pcr_errs',
        scenario='seq_errs',
        # scenario='ampl_rates',
        # scenario='umi_lens',
        # scenario='seq_deps',
        # scenario='umi_nums',

        umi_len=30,

        is_trim=False, # False True
        is_split=True, # False True

        # umi_ref_fpn=to('data/simu/umi/trimer/seq_errs/umi.txt'),
        # fastq_fp=to('data/simu/umi/trimer/seq_errs/trimmed/'),

        param_fpn=to('data/trimer.yml'),

        verbose=True, # False True
    )



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

from umiche.network.Adjacency import Adjacency
from umiche.network.Edge import Edge
from umiche.network.CC import CC


def adjacency(
        graph_adj : Dict,
):
    return Adjacency(
        graph=graph_adj,
    )


def edge(
        graph_edge_list,
):
    return Edge(
        graph=graph_edge_list
    )


def cc(
        graph_adj : Dict,
        method : str = 'deque',
        verbose : bool = True,
):
    if method == "deque":
        return CC(
            verbose=verbose,
        ).deque(
            graph=graph_adj,
        )
    if method == "set":
        return CC(
            verbose=verbose,
        ).set(
            graph=graph_adj,
        )


if __name__ == "__main__":
    graph_adj_umitools = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D'],
        'F': ['D'],
    }
    graph_adj_mclumi = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }

    # adj = adjacency(
    #     # graph_adj=graph_adj_umitools,
    #     graph_adj=graph_adj_mclumi,
    # )
    # print(adj.graph_mapped)
    # print(adj.graph)
    # print(adj.dict())
    # print(adj.set())
    # print(adj.list())
    # print(adj.to_matrix())
    # print(adj.to_edge_list(rr=False))
    # print(adj.to_edge_list(rr=True))

    # eg = edge(
    #     graph_edge_list=adj.to_edge_list(rr=False),
    # )
    # print(eg.graph)
    # print(eg.nodes)
    # print(eg.key_mapped)
    # print(eg.rvredanduncy)
    # eg.graph = eg.rvredanduncy
    # print(eg.graph)
    # print(eg.key_mapped)
    # print(eg.rvredanduncy)
    # print(eg.to_adj_dict())
    # print(eg.graph_mapped)

    connected_components = cc(
        graph_adj=graph_adj_mclumi,
        method='deque',
        verbose=True
    )
    print(list(connected_components))

    connected_components = cc(
        graph_adj=graph_adj_mclumi,
        method='set',
        verbose=True
    )
    print(connected_components)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Set

import itertools
import collections
from umiche.util.Number import Number as rannum
from umiche.util.Console import Console


class Collapse:

    def __init__(
            self,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def majority_vote(
            self,
            umi,
            recur_len=3,
    ) -> str:
        """

        Notes
        --------
            elif len(s) == 2:
                sdict = {umi_trimer.count(i): i for i in s}
                t.append(sdict[2])

        Parameters
        ----------
        umi : str
            an umi CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC
        recur_len : int
            the length of a building block consisting of recurring nucleotides, e.g. trimer 3

        Returns
        -------
        str
            an umi, in which multipolymer nucleotides collapsed into a single nucleotide
        """
        scales = [i for i in range(len(umi)) if i%recur_len == 0]
        umi_blocks = [umi[v: v+recur_len] for v in scales]
        combos = []
        for umi_block in umi_blocks:
            s = set(umi_block)
            if len(s) == recur_len:
                rand_index = rannum().uniform(low=0, high=recur_len, num=1, use_seed=False, seed=3)[0]
                combos.append(umi_block[rand_index])
            elif len(s) == int(recur_len/2) + 1:
                sdict = {base: umi_block.count(base) for base in s}
                combos.append(max(sdict, key=sdict.get))
            else:
                combos.append(umi_block[0])
        return ''.join(combos)

    def take_by_order(
            self,
            umi,
            pos=0,
            recur_len=3,
    ):
        scales = [i for i in range(len(umi)) if i % recur_len == 0]
        umi_blocks = [umi[v: v + recur_len] for v in scales]
        combos = []
        for umi_block in umi_blocks:
            combos.append(umi_block[pos])
        return ''.join(combos)

    def majority_vote_most_common_deprecated(
            self,
            umi,
    ):
        from collections import Counter
        vernier = [i for i in range(36) if i % 3 == 0]
        umi_trimers = [umi[v: v+3] for v in vernier]
        # @@ textwrap is a bit slower than Python list.
        # import textwrap
        # umi_trimers = textwrap.wrap(umi, 3)
        t = []
        for umi_trimer in umi_trimers:
            s = Counter(umi_trimer).most_common()
            if len(s) == 3:
                rand_index = rannum().uniform(low=0, high=3, num=1, use_seed=False)[0]
                t.append(s[rand_index][0])
            elif len(s) == 2:
                t.append(s[0][0])
            else:
                t.append(umi_trimer[0])
        return ''.join(t)

    def split_to_all(
            self,
            umi,
            recur_len=3,
        ) -> Set:
        """

        Parameters
        ----------
        umi : str
            an umi CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC
        recur_len : int
            the length of a building block consisting of recurring nucleotides, e.g. trimer 3

        Returns
        -------
        Set
            a set containing umis, each of which is a nucleotide polymer (trinucleotides)
             collapsed into a single nucleotide.
            e.g. {'TCATCTAGTGCC', 'TGCTCTAGTGCC', 'CGTTCTAGTGCC', ..., 'TCCTCTAGTGCC'}

        """
        # @@ umi
        # CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC
        scales = [i for i in range(len(umi)) if i % recur_len == 0]
        # @@ scales
        # [0, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33]
        umi_blocks = [umi[v: v + recur_len] for v in scales]
        # @@ umi_blocks
        # ['CTT', 'CCG', 'CAT', 'TTT', 'CCC', 'TTT', 'AAA', 'GGG', 'TTT', 'GGG', 'CCC', 'CCC']
        combos = [umi_block for umi_block in set(umi_blocks[0])]
        # @@ combos
        # ['C', 'T']
        umi_blocks = umi_blocks[1:]
        # @@ umi_blocks
        # ['CCG', 'CAT', 'TTT', 'CCC', 'TTT', 'AAA', 'GGG', 'TTT', 'GGG', 'CCC', 'CCC']
        for id, umi_block in enumerate(umi_blocks):
            umi_block_bases = set(umi_block)
            c = []
            for i in range(len(combos)):
                for umi_block_base in umi_block_bases:
                    c.append(combos[i] + umi_block_base)
            combos = c
            # @@ combos
            # ['CC', 'CG', 'TC', 'TG']
            # ['CCC', 'CCT', 'CCA', 'CGC', 'CGT', 'CGA', 'TCC', 'TCT', 'TCA', 'TGC', 'TGT', 'TGA']
            # ['CCCT', 'CCTT', 'CCAT', 'CGCT', 'CGTT', 'CGAT', 'TCCT', 'TCTT', 'TCAT', 'TGCT', 'TGTT', 'TGAT']
            # ['CCCTC', 'CCTTC', 'CCATC', 'CGCTC', 'CGTTC', 'CGATC', 'TCCTC', 'TCTTC', 'TCATC', 'TGCTC', 'TGTTC', 'TGATC']
            # ['CCCTCT', 'CCTTCT', 'CCATCT', 'CGCTCT', 'CGTTCT', 'CGATCT', 'TCCTCT', 'TCTTCT', 'TCATCT', 'TGCTCT', 'TGTTCT', 'TGATCT']
            # ['CCCTCTA', 'CCTTCTA', 'CCATCTA', 'CGCTCTA', 'CGTTCTA', 'CGATCTA', 'TCCTCTA', 'TCTTCTA', 'TCATCTA', 'TGCTCTA', 'TGTTCTA', 'TGATCTA']
            # ['CCCTCTAG', 'CCTTCTAG', 'CCATCTAG', 'CGCTCTAG', 'CGTTCTAG', 'CGATCTAG', 'TCCTCTAG', 'TCTTCTAG', 'TCATCTAG', 'TGCTCTAG', 'TGTTCTAG', 'TGATCTAG']
            # ['CCCTCTAGT', 'CCTTCTAGT', 'CCATCTAGT', 'CGCTCTAGT', 'CGTTCTAGT', 'CGATCTAGT', 'TCCTCTAGT', 'TCTTCTAGT', 'TCATCTAGT', 'TGCTCTAGT', 'TGTTCTAGT', 'TGATCTAGT']
            # ['CCCTCTAGTG', 'CCTTCTAGTG', 'CCATCTAGTG', 'CGCTCTAGTG', 'CGTTCTAGTG', 'CGATCTAGTG', 'TCCTCTAGTG', 'TCTTCTAGTG', 'TCATCTAGTG', 'TGCTCTAGTG', 'TGTTCTAGTG', 'TGATCTAGTG']
            # ['CCCTCTAGTGC', 'CCTTCTAGTGC', 'CCATCTAGTGC', 'CGCTCTAGTGC', 'CGTTCTAGTGC', 'CGATCTAGTGC', 'TCCTCTAGTGC', 'TCTTCTAGTGC', 'TCATCTAGTGC', 'TGCTCTAGTGC', 'TGTTCTAGTGC', 'TGATCTAGTGC']
            # ['CCCTCTAGTGCC', 'CCTTCTAGTGCC', 'CCATCTAGTGCC', 'CGCTCTAGTGCC', 'CGTTCTAGTGCC', 'CGATCTAGTGCC', 'TCCTCTAGTGCC', 'TCTTCTAGTGCC', 'TCATCTAGTGCC', 'TGCTCTAGTGCC', 'TGTTCTAGTGCC', 'TGATCTAGTGCC']
        # @@ set(combos)
        # {'TCTTCTAGTGCC', 'CGATCTAGTGCC', 'CGTTCTAGTGCC', 'CCCTCTAGTGCC', 'TGATCTAGTGCC', 'CCATCTAGTGCC', 'CCTTCTAGTGCC', 'TGCTCTAGTGCC', 'TCATCTAGTGCC', 'TCCTCTAGTGCC', 'CGCTCTAGTGCC', 'TGTTCTAGTGCC'}
        return set(combos)

    def split_by_mv(
            self,
            umi,
            recur_len=3,
    ) -> Set:
        """

        Parameters
        ----------
        umi : str
            an umi CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC
        recur_len : int
            the length of a building block consisting of recurring nucleotides, e.g. trimer 3

        Returns
        -------
        Set
            a set containing umis, each of which is a nucleotide polymer (trinucleotides)
             collapsed into a single nucleotide.
            e.g. {'TCATCTAGTGCC', 'TCTTCTAGTGCC', 'TCCTCTAGTGCC'}

        """
        scales = [i for i in range(len(umi)) if i % recur_len == 0]
        # @@ scales
        # [0, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33]
        umi_blocks = [umi[v: v + recur_len] for v in scales]
        # @@ umi_blocks
        # ['CTT', 'CCG', 'CAT', 'TTT', 'CCC', 'TTT', 'AAA', 'GGG', 'TTT', 'GGG', 'CCC', 'CCC']
        umi_block_bases = self.vote(umi_blocks[0], recur_len)
        # @@ umi_block_bases
        # {'T'}
        combos = [umi_block for umi_block in umi_block_bases]
        # @@ combos
        # ['T']
        umi_blocks = umi_blocks[1:]
        for id, umi_block in enumerate(umi_blocks):
            umi_block_bases = self.vote(umi_block, recur_len)
            # @@ umi_block_bases
            # {'C'}
            # {'T', 'A', 'C'}
            # {'T'}
            # {'C'}
            # {'T'}
            # {'A'}
            # {'G'}
            # {'T'}
            # {'G'}
            # {'C'}
            # {'C'}
            c = []
            for i in range(len(combos)):
                for umi_block_base in umi_block_bases:
                    c.append(combos[i] + umi_block_base)
            combos = c
            # @@ combos
            # ['TC']
            # ['TCC', 'TCT', 'TCA']
            # ['TCCT', 'TCTT', 'TCAT']
            # ['TCCTC', 'TCTTC', 'TCATC']
            # ['TCCTCT', 'TCTTCT', 'TCATCT']
            # ['TCCTCTA', 'TCTTCTA', 'TCATCTA']
            # ['TCCTCTAG', 'TCTTCTAG', 'TCATCTAG']
            # ['TCCTCTAGT', 'TCTTCTAGT', 'TCATCTAGT']
            # ['TCCTCTAGTG', 'TCTTCTAGTG', 'TCATCTAGTG']
            # ['TCCTCTAGTGC', 'TCTTCTAGTGC', 'TCATCTAGTGC']
            # ['TCCTCTAGTGCC', 'TCTTCTAGTGCC', 'TCATCTAGTGCC']
        return set(combos)

    def vote(
            self,
            umi_block,
            recur_len,
    ) -> Set:
        """
        It makes a vote in a umi block by choosing the most common item.

        Parameters
        ----------
        umi_block : str
            a umi building block, e.g., homo-trinucleotide
        recur_len : int
            the length of a building block consisting of recurring nucleotides, e.g. trimer 3

        Returns
        -------
        Set
            unrepeated cleotides, e.g., {'T', 'A', 'C'}
        """
        if len(set(umi_block)) == recur_len:
            return set(umi_block)
        elif len(set(umi_block)) == int(recur_len / 2) + 1:
            sdict = {base: umi_block.count(base) for base in set(umi_block)}
            return set(max(sdict, key=sdict.get))
        else:
            return set(umi_block[0])

    def umi_li(
            self,
            umi,
    ) -> Set:
        """

        Parameters
        ----------
        umi : str
            an umi CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC

        Returns
        -------
        Set
            a set containing umis, each of which is a nucleotide polymer (trinucleotides)
             collapsed into a single nucleotide.
            e.g. {'TCATCTAGTGCC', 'TGCTCTAGTGCC', 'CGTTCTAGTGCC', ..., 'TCCTCTAGTGCC'}

        """
        trimers = list()
        umis = set()
        while len(umi) != 0:
            try:
                trimers.append(set(umi[0:3]))
                umi = umi[3:]
            except:
                self.console.print("umi existing indel or wrong")
        for val in itertools.product(*trimers):
            collapsed_umi = ''.join(val)
            umis.add(collapsed_umi)
        return umis

    def cmi_li(
            self,
            umi,
    ) -> Set:
        """
        This implements nu collapsing each trimer block into one
        by Canzar's lab (Stefan Canzar, Shuang Li, and Pablo Monteagudo-Mesas).

        Notes
        -----
        This implementation is the same as split_by_mv() implemented by Jianfeng Sun.

        Parameters
        ----------
        umi : str
            an umi CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC

        Returns
        -------
        Set
            a set containing umis, each of which is a nucleotide polymer (trinucleotides)
             collapsed into a single nucleotide.
            e.g. {'TCATCTAGTGCC', 'TCTTCTAGTGCC', 'TCCTCTAGTGCC'}

        """
        trimers = list()
        umis = set()
        while len(umi) != 0:
            try:
                if len(set(umi[0:3])) < 3:
                    base = collections.Counter(umi[0:3]).most_common(1)[0][0]
                    trimers.append(base)
                    umi = umi[3:]
                else:
                    trimers.append(set(umi[0:3]))
                    umi = umi[3:]
            except:
                self.console.print("umi existing indel or wrong")
        for val in itertools.product(*trimers):
            collapse = ''.join(val)
            umis.add(collapse)
        return umis


if __name__ == "__main__":
    p = Collapse()
    # set1 = p.umi_li('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC')
    # set2 = p.split_to_all('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3)
    # print(set1)
    # print(set2)
    # print(set1 == set2)

    # set3 = p.cmi_li('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC')
    # set4 = p.split_by_mv('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3)
    # print(set3)
    # print(set4)
    # print(set3 == set4)
    #
    # umi_collapsed = p.majority_vote('CTGCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC') # T1TTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC
    # print(umi_collapsed)

    # print(p.split_by_mv('GGTTAAGGAAGGTTGGAAGTAATT', recur_len=2))
    # print(p.split_to_all('GGTTAAGGAAGGTTGGAAGTAATT', recur_len=2))
    # print(p.majority_vote('GGTTAAGGAAGGTTGGAAGTAATT', recur_len=2))


    # @@ UMIche work Figure 2
    print(p.split_by_mv('AAATCCGGATTTCGGAAATTTGGGCACCCC', recur_len=3))

    # print(p.split_to_all('AAATCCGGATTTCGGAAATTTGGGCACCCC', recur_len=3))
    # print(p.split_to_all('AAATCCGGATTTCGGAAATTTGGGCCCCCC', recur_len=3))
    # print(p.split_to_all('AAATCCGGATTTGGGAAATTTGGGCCCCCC', recur_len=3))
    # print(p.split_to_all('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))
    #
    # print(p.majority_vote('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))
    # a1 = p.split_to_all('AAATCCGGATTTCGGAAATTTGGGCACCCC', recur_len=3)
    # a2 = p.split_to_all('AAATCCGGATTTGGGAAACTTGGGCACCCC', recur_len=3)
    # a3 = p.split_to_all('AAATCCGGATTTCGGAAATTTGGGCCCCCC', recur_len=3)
    # print(a2)
    # print(a1 - a2)
    # print(a1 - a3)
    # print(a2 - a3)

    # set4 = p.split_by_mv('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3)
    # set2 = p.split_to_all('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3)
    # print(p.split_to_all('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3))
    # print(p.split_by_mv('CTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3))

    # print(p.split_to_all('TTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3))
    # print(p.split_by_mv('TTTCCGCATTTTCCCTTTAAAGGGTTTGGGCCCCCC', recur_len=3))



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"

from typing import Union
import numpy as np

from umiche.util.Console import Console


class TMR:
    
    def __init__(
            self,
            error_rate: Union[float, np.ndarray] = 0.00001,
            num_nt: int = 12,
            verbose=True,
    ):
        self.error_rate = error_rate
        self._num_nt = num_nt
        # self.bitflip_block_error = 3 * self.error_rate**2 - 2 * self.error_rate**3
        # self.homotrimer_block_error = (7/3) * self.error_rate**2 * (1 - self.error_rate) + self.error_rate**3

        self.verbose = verbose
        self.console = Console()
        self.console.verbose = self.verbose

    def get_num_nt(self):
        self.console.print("=========>number of building blocks in a UMI is {}".format(self._num_nt))
        return self._num_nt

    def set_num_nt(self, value):
        # self.console.print("=========>number of building blocks in a UMI is set to {}:".format(value))
        self._num_nt = value

    @property
    def homotrimer_block_error(self, ):
        e = (7/3) * self.error_rate**2 * (1 - self.error_rate) + self.error_rate**3
        self.console.print("=========>homotrimer block error rate: {}".format(e))
        return e

    @property
    def bitflip_block_error(self, ):
        e = 3 * self.error_rate ** 2 - 2 * self.error_rate ** 3
        self.console.print("=========>binary repetition code block error rate is {}".format(e))
        return e

    @property
    def homotrimer_block_error_free(self, ):
        e = (7 / 3) * self.error_rate ** 2 * (1 - self.error_rate) + self.error_rate ** 3
        self.console.print("=========>homotrimer block error rate: {}".format(e))
        return 1 - e

    @property
    def bitflip_block_error_free(self, ):
        e = 3 * self.error_rate ** 2 - 2 * self.error_rate ** 3
        self.console.print("=========>binary repetition code block error rate is {}".format(e))
        return 1 - e

    @property
    def monomer_umi_error(self, ):
        self.console.print("======>monomer UMI error rate:")
        return 1 - (1 - self.error_rate) ** self.get_num_nt()

    @property
    def homotrimer_umi_error(self, ):
        self.console.print("======>homotrimer UMI error rate:")
        return 1 - (1 - self.homotrimer_block_error) ** self.get_num_nt()

    @property
    def bitflip_code_error(self, ):
        self.console.print("======>binary repetition code error rate:")
        return 1 - (1 - self.bitflip_block_error) ** self.get_num_nt()

    @property
    def monomer_umi_error_free(self, ):
        self.console.print("======>monomer UMI successful synthesis rate:")
        return (1 - self.error_rate) ** self.get_num_nt()

    @property
    def homotrimer_umi_error_free(self, ):
        self.console.print("======>homotrimer UMI successful synthesis rate:")
        return (1 - self.homotrimer_block_error) ** self.get_num_nt()

    @property
    def bitflip_code_error_free(self, ):
        self.console.print("======>binary repetition code successful transmission rate:")
        return (1 - self.bitflip_block_error) ** self.get_num_nt()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd


def read(
        gmat_fpn,
):
    df_gmat = pd.read_hdf(gmat_fpn, 'df')
    y = df_gmat['Y'].values
    if 'Y' in df_gmat.columns:
        df_gmat = df_gmat.drop(columns=['Y'])
    return df_gmat, y


df, y = read(
    gmat_fpn=to('mclumi/gmat_customized.h5'),
)

print(df)

condi_map ={
    'umi_nums_seq_err_0.1': 'Sequencing error 0.1',
    'umi_nums_seq_err_0.05': 'Sequencing error 0.05',
    'umi_nums_seq_err_0.01': 'Sequencing error 0.01',
    'umi_nums_seq_err_0.005': 'Sequencing error 0.005',
}

condi = 'umi_nums_seq_err_0.1'
# condi = 'umi_nums_seq_err_0.05'
# condi = 'umi_nums_seq_err_0.01'
# condi = 'umi_nums_seq_err_0.005'

direc_dedup = pd.read_csv(to('mclumi/') + condi + '/directional_dedup.txt', header=0)
mcl_dedup = pd.read_csv(to('mclumi/') + condi + '/mcl_dedup.txt', header=0)
# print(direc_dedup)
# print(mcl_dedup)

# print(np.unique(df.to_numpy().flatten()).tolist())
uniq = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45]

dedup_direc_map = {x:y for x, y in zip(uniq, direc_dedup['pn0'].values.tolist())}
dedup_mcl_map = {x:y for x, y in zip(uniq, mcl_dedup['pn0'].values.tolist())}
print(dedup_direc_map)
print(dedup_mcl_map)

x = df.values
x_direc = df.replace(uniq, direc_dedup['pn0'].values.tolist()).values
x_mcl = df.replace(uniq, mcl_dedup['pn0'].values.tolist()).values

from ext.sequelpy.plot.scatter.DimensionReduction import dimensionReduction as drplot # TSNE
# drplot().single(X=x, y=y, tech='TSNE', marker_size=3, cmap='Paired', title='Real')
# drplot().single(X=x_direc, y=y, tech='TSNE', marker_size=3, cmap='Paired', title='Directional' + ' (' +  condi_map[condi] + ')')
# drplot().single(X=x_mcl, y=y, tech='TSNE', marker_size=3, cmap='Paired', title='MCL' + ' (' + condi_map[condi] + ')')
# drplot().single(X=xs, y=ys, tech='PCA', marker_size=3, cmap='tab20b', title='CondiCVAE')
# drplot().single(X=xs, y=ys, tech='UMAP', marker_size=3, cmap='tab20b', title='CondiCVAE')

drplot().single(X=x_direc, y=y, tech='UMAP', marker_size=3, cmap='Paired', title='Directional' + ' (' +  condi_map[condi] + ')')
drplot().single(X=x_mcl, y=y, tech='UMAP', marker_size=3, cmap='Paired', title='MCL' + ' (' + condi_map[condi] + ')')


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import pandas as pd
from umiche.util.Console import Console


class Reader:

    def __init__(
            self,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose

    def todf(self, seqs, names):
        umi_df_stime = time.time()
        df_fastq = pd.DataFrame(seqs, columns=['seq_raw'])
        df_fastq['name'] = names
        df_fastq['umi'] = df_fastq['name'].apply(lambda x: x.split('_')[1])
        df_fastq['umi#'] = df_fastq['name'].apply(lambda x: x.split('_')[0].split('-')[0])
        df_fastq['umi_src'] = df_fastq['name'].apply(lambda x: x.split('_')[0].split('-')[1])
        # df_fastq['umi_pcr#'] = df_fastq['name'].apply(lambda x: self.pcrnum(x))
        df_fastq['umi#'] = df_fastq['umi#'].astype(int)
        # print(df_fastq)
        # print(df_fastq['name'])
        # print(df_fastq['umi'])
        # print(df_fastq['umi#'])
        # print(df_fastq['umi_src'])
        self.console.print('===============>Trimmed UMIs to df time: {:.3f}s'.format(time.time() - umi_df_stime))
        return df_fastq

    def todfFromTree(self, seqs, names):
        import pandas as pd
        umi_df_stime = time.time()
        df_fastq = pd.DataFrame(seqs, columns=['seq_raw'])
        df_fastq['name'] = names
        df_fastq['umi'] = df_fastq['name'].apply(lambda x: x.split('_')[-1])
        df_fastq['umi#'] = df_fastq['name'].apply(lambda x: self.srcnumTickout(x))
        df_fastq['umi_src'] = df_fastq['name'].apply(lambda x: '-'.join(x.split('_')[1:-1]))
        # print(df_fastq)
        # df_fastq['umi_pcr#'] = df_fastq['name'].apply(lambda x: self.pcrnum(x))
        df_fastq['umi#'] = df_fastq['umi#'].astype(int)
        # print(df_fastq)
        # print(df_fastq['name'])
        # print(df_fastq['umi'])
        # print(df_fastq['umi#'])
        # print(df_fastq['umi_src'])
        self.console.print('===>Trimmed UMIs to df time: {:.3f}s'.format(time.time() - umi_df_stime))
        return df_fastq

    def pcrnum(self, x):
        c = x.split('_')[0].split('-')
        if c[1] == 'init':
            return -1
        else:
            return c[2]

    def srcnumTickout(self, x):
        val = x.split('_')[0]
        return val.split('-')[0]




class Style:

    def __init__(self, ):
        pass

    def deephelicon(self, boxplot_handles, palette=[]):
        """

        ..  Description:
            ------------
            1. bplot_handle offers six properties to adjust, including:
            1) 'whiskers', 2) 'caps', 3) 'medians', 4) 'fliers', 5) 'means', 6) 'boxes'.

            2. style scheme used in Figure S1 in the DeepHelicon (link 1) paper.

        ..  See:
            ----
            1. DeepHelicon: https://doi.org/10.1016/j.jsb.2020.107574.

        :param boxplot_handles: dict - plt boxplot obs
        :param palette: 1d list
        :return:
        """
        for bplot_handle, color in zip(boxplot_handles, palette):
            for patch in bplot_handle['boxes']:
                patch.set(color='black', linewidth=1.1)
                patch.set(facecolor='white', alpha=0.3)
            for whisker in bplot_handle['whiskers']:
                whisker.set(color='black', linewidth=1)
            for cap in bplot_handle['caps']:
                cap.set(color='black', linewidth=1)
            for median in bplot_handle['medians']:
                median.set(color='black', linewidth=1)
            for flier in bplot_handle['fliers']:
                flier.set(marker='o', color='y', alpha=0.5)
        return

    def plain(self, boxplot_handles, palette, edge_width=1):
        """

        ..  Description:
            ------------
            bplot_handle offers six properties to adjust, including:
            1) 'whiskers', 2) 'caps', 3) 'medians', 4) 'fliers', 5) 'means', 6) 'boxes'.


        :param boxplot_handles: dict - plt boxplot obs
        :param palette: 1d list
        :param edge_width:
        :return:
        """
        for bplot_handle, color in zip(boxplot_handles, palette):
            for patch in bplot_handle['boxes']:
                # b.set_alpha(0.6)
                # patch.set_edgecolor(color, alpha=0.3)
                patch.set(
                    edgecolor=color,
                    alpha=0.7,
                )
                # patch.set_facecolor('white')
                patch.set_linewidth(edge_width)
                patch.set(
                    facecolor='white'
                )
            for whisker in bplot_handle['whiskers']:
                whisker.set(
                    color=color,
                    linewidth=2,
                    linestyle='dashed', # the same as '--'
                    alpha=0.5,
                )
            for cap in bplot_handle['caps']:
                cap.set(color=color, linewidth=1)
            for median in bplot_handle['medians']:
                median.set(color=color, linewidth=1)
            for flier in bplot_handle['fliers']:
                flier.set(marker='o', color='y', alpha=0.5)

    def plaincolor(self, plt, boxplot_handles, palette=[]):
        """
        ..  Description:
            ------------
            only adjust uniform colors each boxplot.

        :param plt:
        :param boxplot_handles: dict - plt boxplot obs
        :param palette: 1d list
        :return:
        """
        def set(bp_handle, color):
            plt.setp(bp_handle['boxes'], color=color)
            plt.setp(bp_handle['whiskers'], color=color)
            plt.setp(bp_handle['caps'], color=color)
            plt.setp(bp_handle['medians'], color=color)
        for i_bp, c in zip(boxplot_handles, palette):
            set(i_bp, c)
        return


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import networkx as nx
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.plot.gadget.Element import Element as pele
from umiche.network.Adjacency import Adjacency as netadj


class Graph:

    def __init__(
            self,
            graph,
            which_color="tableau",
    ):
        self.pele = pele()
        self.netadj = netadj()
        self.graph = graph
        self.netadj.graph = self.graph
        self.color_list = self.pele.color(which=which_color, is_random=True)

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def draw(
            self,
            cc_dict,
            title,
            ax,
    ):
        el = self.netadj.to_edge_list()
        color_per_subcc = {}
        for cc_i, (cc_id, cc) in enumerate(cc_dict.items()):
            for node_j, node in enumerate(cc):
                color_per_subcc[node] = self.color_list[cc_i]
        print(color_per_subcc)
        color_per_subcc = [color_per_subcc[k] for k in self.graph.keys()]

        G = nx.Graph()
        G.add_nodes_from(self.graph.keys())
        G.add_edges_from(el)
        print(G.nodes())

        # pos = {
        #     'A:120': (0, 0),
        #     'B:2': (-1, 1),
        #     'C:2': (1, 1),
        #     'D:90': (0, -2.5),
        #     'E:50': (-1, -4.5),
        #     'F:1': (1, -4.5),
        #     'G:5': (0, -6),
        # }

        pos = {
            'A': (0, 0),
            'B': (-1, 1),
            'C': (1, 1),
            'D': (0, -2.5),
            'E': (-1, -4.5),
            'F': (1, -4.5),
        }

        options = {
            "font_size": 14,
            "node_size": 2000,
            "node_color": "white", # bisque
            "edgecolors": color_per_subcc, # node edge color
            "edge_color": 'dimgray', # simply edge color
            "linewidths": 3,
            "width": 4,
        }
        nx.draw_networkx(G, pos, ax=ax, **options)
        ax.set_title(title, fontsize=18)
        ax.set_axis_off()


if __name__ == "__main__":
    import pandas as pd

    # graph_adj = {
    #     'A:120': ['B:2', 'C:2', 'D:90'],
    #     'B:2': ['A:120', 'C:2'],
    #     'C:2': ['A:120', 'B:2'],
    #     'D:90': ['A:120', 'E:50', 'F:1'],
    #     'E:50': ['D:90', 'G:5'],
    #     'F:1': ['D:90', 'G:5'],
    #     'G:5': ['E:50', 'F:1'],
    # }

    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D', 'G'],
    #     'F': ['D', 'G'],
    #     'G': ['E', 'F'],
    # }
    # print("An adjacency list of a graph:\n{}".format(graph_adj))

    # node_val_sorted = pd.Series({
    #     'A:120': 120,
    #     'D:90': 90,
    #     'E:50': 50,
    #     'G:5': 5,
    #     'B:2': 2,
    #     'C:2': 2,
    #     'F:1': 1,
    # })
    # print("Counts sorted:\n{}".format(node_val_sorted))

    ### @@ data from UMI-tools
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D'],
        'F': ['D'],
    }
    print("An adjacency list of a graph:\n{}".format(graph_adj))

    node_val_sorted = pd.Series({
        'A': 456,
        'E': 90,
        'D': 72,
        'B': 2,
        'C': 2,
        'F': 1,
    })
    print("Counts sorted:\n{}".format(node_val_sorted))

    # ccs = [['A:120', 'B:2', 'C:2', 'D:90', 'E:50', 'F:1']]  # cc
    # ccs = [['A:120', 'B:2', 'C:2'], ['D:90', 'E:50', 'F:1'], ['G:5']]  # adj
    # ccs = [['A:120', 'B:2', 'C:2'], ['D:90', 'F:1', ], ['E:50', 'G:5']]  # direc
    # ccs = [['A:120', 'B:2', 'C:2'], ['D:90', 'E:50', 'F:1', 'G:5']] # mcl
    # ccs = [['A:120', 'B:2', 'C:2', 'D:90', 'E:50', 'F:1', 'G:5']] # mcl_ed
    # ccs = [['A:120', 'B:2', 'C:2'], ['D:90', 'E:50', 'F:1', 'G:5']] # mcl_val

    ### @@@ ******Connected components******
    from umiche.deduplicate.method.Cluster import Cluster as umiclust
    ccs = umiclust().cc(graph_adj=graph_adj)
    print("Connected components:\n{}".format(ccs))

    ### @@@ ******UMI-tools Adjacency******
    from umiche.deduplicate.method.Adjacency import Adjacency as umiadj
    from umiche.deduplicate.method.Directional import Directional as umidirec
    from umiche.deduplicate.method.MarkovClustering import MarkovClustering as umimcl
    dedup_res_adj = umiadj().umi_tools(
        connected_components=ccs,
        df_umi_uniq_val_cnt=node_val_sorted,
        graph_adj=graph_adj,
    )
    dedup_res_adj_dc = umiadj().decompose(dedup_res_adj['clusters'])
    print("deduplicated clusters (UMI-tools Adjacency):\n{}".format(dedup_res_adj_dc))

    ### @@@ ******UMI-tools Directional******
    from umiche.deduplicate.method.Directional import Directional as umidirec
    from umiche.deduplicate.method.MarkovClustering import MarkovClustering as umimcl
    dedup_res_direc = umidirec().umi_tools(
        connected_components=ccs,
        df_umi_uniq_val_cnt=node_val_sorted,
        graph_adj=graph_adj,
    )
    dedup_res_direc_dc = umidirec().decompose(dedup_res_direc['clusters'])
    print("deduplicated clusters (UMI-tools Directional):\n{}".format(dedup_res_direc_dc))

    ### @@@ ******MCL******
    from umiche.deduplicate.method.MarkovClustering import MarkovClustering as umimcl
    mcl = umimcl(
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
    )
    df_mcl = mcl.dfclusters(
        connected_components=ccs,
        graph_adj=graph_adj,
    )
    dedup_res_mcl_dc = mcl.decompose(list_nd=df_mcl['clusters'].values)
    print("deduplicated clusters (MCL):\n{}".format(dedup_res_mcl_dc))

    ### @@@ ******MCL mcl_val******
    df_mcl_val = mcl.maxval_val(
        df_mcl_ccs=df_mcl,
        df_umi_uniq_val_cnt=node_val_sorted,
        thres_fold=2,
    )
    dedup_res_mcl_val_dc = mcl.decompose(list_nd=df_mcl_val['clusters'].values)
    print("deduplicated clusters decomposed (mcl_val):\n{}".format(dedup_res_mcl_val_dc))
    dedup_res_mcl_val_dc_full = mcl.get_full_subcc(ccs_dict=dedup_res_mcl_val_dc, mcl_ccs_dict=dedup_res_mcl_dc)
    print("deduplicated clusters decomposed full list(mcl_val):\n{}".format(dedup_res_mcl_val_dc_full))

    ### @@@ ******MCL mcl_ed******
    # int_to_umi_dict = {
    #     'A:120': 'AGATCTCGCA',
    #     'B:2': 'AGATCCCGCA',
    #     'C:2': 'AGATCACGCA',
    #     'D:90': 'AGATCGCGCA',
    #     'E:50': 'AGATCGCGGA',
    #     'F:1': 'AGATCGCGTA',
    #     'G:5': 'TGATCGCGAA',
    # }
    int_to_umi_dict = {
        'A': 'ACGT',
        'B': 'TCGT',
        'C': 'CCGT',
        'D': 'ACAT',
        'E': 'ACAG',
        'F': 'AAAT',
    }
    df_mcl_ed = mcl.maxval_ed(
        df_mcl_ccs=df_mcl,
        df_umi_uniq_val_cnt=node_val_sorted,
        thres_fold=1,
        int_to_umi_dict=int_to_umi_dict,
    )
    dedup_res_mcl_ed_dc = mcl.decompose(list_nd=df_mcl_ed['clusters'].values)
    print("deduplicated clusters decomposed (mcl_ed):\n{}".format(dedup_res_mcl_ed_dc))
    dedup_res_mcl_ed_dc_full = mcl.get_full_subcc(ccs_dict=dedup_res_mcl_ed_dc, mcl_ccs_dict=dedup_res_mcl_dc)
    print("deduplicated clusters decomposed full list(mcl_ed):\n{}".format(dedup_res_mcl_ed_dc_full))

    fig, ax = plt.subplots(nrows=2, ncols=3, figsize=(14, 9.5))
    p = Graph(graph=graph_adj)
    p.color_list = ['cornflowerblue', 'lightcoral', 'mediumseagreen',]
    p.draw(ccs, ax=ax[0, 0], title='Cluster')
    p.draw(dedup_res_adj_dc, title='Adjacent', ax=ax[0, 1])
    p.draw(dedup_res_direc_dc, title='Directional', ax=ax[0, 2])
    p.draw(dedup_res_mcl_dc, title='MCL', ax=ax[1, 0])
    p.draw(dedup_res_mcl_val_dc_full, title='MCL-val', ax=ax[1, 1])
    p.draw(dedup_res_mcl_ed_dc_full, title='MCL-ed', ax=ax[1, 2])

    plt.subplots_adjust(
        top=0.92,
        bottom=0.04,
        left=0.04,
        right=0.98,
        hspace=0.10,
        wspace=0.15,
    )
    plt.show()



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from functools import wraps


class Reader:



    def __call__(self, deal):
        generic = self.generic
        excel = self.excel
        @wraps(deal)
        def read(ph, *args, **kwargs):
            deal(ph, **kwargs)
            keys = [*kwargs.keys()]
            if kwargs['type'] == 'generic':
                return generic(
                    df_fpn=kwargs['df_fpn'],
                    df_sep='\t' if 'df_sep' not in keys else kwargs['df_sep'],
                    skiprows=False if 'skiprows' not in keys else kwargs['skiprows'],
                    header=None if 'header' not in keys else kwargs['header'],
                    is_utf8=False if 'is_utf8' not in keys else kwargs['is_utf8'],
                )
            elif kwargs['type'] == 'excel':
                return excel(
                    df_fpn=kwargs['df_fpn'],
                    sheet_name='Sheet1' if 'sheet_name' not in keys else kwargs['sheet_name'],
                    header=None if 'header' not in keys else kwargs['header'],
                    is_utf8=False if 'is_utf8' not in keys else kwargs['is_utf8'],
                )
        return read

    def generic(self, df_fpn, df_sep='\t', skiprows=None, header=None, index_col=None, is_utf8=False):
        """

        Parameters
        ----------
        df_fpn
        df_sep
        skiprows
        header
        is_utf8

        Returns
        -------

        """
        if is_utf8:
            return pd.read_csv(
                df_fpn,
                sep=df_sep,
                header=header,
                index_col=index_col,
                encoding='utf-8',
                skiprows=skiprows,
            )
        else:
            return pd.read_csv(
                df_fpn,
                sep=df_sep,
                header=header,
                index_col=index_col,
                skiprows=skiprows
            )

    def excel(self, df_fpn, sheet_name='Sheet1', header=None, is_utf8=False):
        """

        Parameters
        ----------
        df_fpn
        sheet_name
        header
        is_utf8

        Returns
        -------

        """
        if is_utf8:
            return pd.read_excel(
                df_fpn,
                sheet_name=sheet_name,
                header=header,
                encoding='utf-8',
                engine='openpyxl',
            )
        else:
            return pd.read_excel(
                df_fpn,
                sheet_name=sheet_name,
                header=header,
                engine='openpyxl',
            )


__author__ = "Jianfeng Sun"
__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL v3.0"
__email__ = "jianfeng.sunmt@gmail.com"
__maintainer__ = "Jianfeng Sun"

from functools import wraps
from matplotlib.ticker import FuncFormatter


class Draw:

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.type = self.kwargs['type']

    def __call__(self, deal):
        if self.type == 'bar_individual':
            draw = self.bar_individual
        # if self.type == 'bar_grouped':
        #     draw = self.bar_grouped
        if self.type == 'hist_grouped':
            draw = self.hist_grouped
        if self.type == 'line':
            draw = self.line
        if self.type == 'line_scatter':
            draw = self.line_scatter

        @wraps(deal)
        def config(self, *args, **kwargs):
            params_all = deal(self, *args, **kwargs)
            # print(kwargs)
            # print(params_all)
            kwarg_keys = kwargs.keys()
            if "color" not in kwarg_keys:
                color = None
            else:
                if kwargs['color'] == "by default":
                    color = 'grey'
                else:
                    color = kwargs['color']

            if "edgecolor" not in kwarg_keys:
                edgecolor = None
            else:
                if kwargs['edgecolor'] == "by default":
                    edgecolor = 'grey'
                else:
                    edgecolor = kwargs['edgecolor']

            if "facecolor" not in kwarg_keys:
                facecolor = None
            else:
                if kwargs['facecolor'] == "by default":
                    facecolor = 'grey'
                else:
                    facecolor = kwargs['facecolor']

            if "width" not in kwarg_keys:
                width = 0.8
            else:
                if kwargs['width'] == "by default":
                    width = 0.4
                else:
                    width = kwargs['width']

            if "linewidth" not in kwarg_keys:
                linewidth = None
            else:
                if kwargs['linewidth'] == "by default":
                    linewidth = 1
                else:
                    linewidth = kwargs['linewidth']

            if "alpha" not in kwarg_keys:
                alpha = None
            else:
                if kwargs['alpha'] == "by default":
                    alpha = 1
                else:
                    alpha = kwargs['alpha']

            if "label" not in kwarg_keys:
                label = None
            else:
                if kwargs['label'] == "by default":
                    label = ""
                else:
                    label = kwargs['label']

            if "x_label" not in kwarg_keys:
                x_label = None
            else:
                if kwargs['x_label'] == "by default":
                    x_label = ""
                else:
                    x_label = kwargs['x_label']

            if "y_label" not in kwarg_keys:
                y_label = None
            else:
                if kwargs['y_label'] == "by default":
                    y_label = ""
                else:
                    y_label = kwargs['y_label']

            if "title" not in kwarg_keys:
                title = None
            else:
                if kwargs['title'] == "by default":
                    title = ""
                else:
                    title = kwargs['title']

            if "x_ticks" not in kwarg_keys:
                x_ticks = None
            else:
                if kwargs['x_ticks'] == "by default":
                    x_ticks = None
                else:
                    x_ticks = kwargs['x_ticks']

            if "y_ticks" not in kwarg_keys:
                y_ticks = None
            else:
                if kwargs['y_ticks'] == "by default":
                    y_ticks = None
                else:
                    y_ticks = kwargs['y_ticks']

            if "x_ticklabels" not in kwarg_keys:
                x_ticklabels = None
            else:
                if kwargs['x_ticklabels'] == "by default":
                    x_ticklabels = None
                else:
                    x_ticklabels = kwargs['x_ticklabels']

            if "x_label_rotation" not in kwarg_keys:
                x_label_rotation = 0
            else:
                if kwargs['x_label_rotation'] == "by default":
                    x_label_rotation = 0
                else:
                    x_label_rotation = kwargs['x_label_rotation']

            if "x_label_rotation_align" not in kwarg_keys:
                x_label_rotation_align = 'center'
            else:
                if kwargs['x_label_rotation_align'] == "by default":
                    x_label_rotation_align = 'center'
                else:
                    x_label_rotation_align = kwargs['x_label_rotation_align']

            if "y_ticklabels" not in kwarg_keys:
                y_ticklabels = None
            else:
                if kwargs['y_ticklabels'] == "by default":
                    y_ticklabels = None
                else:
                    y_ticklabels = kwargs['y_ticklabels']

            if "x_ticklabel_fs" not in kwarg_keys:
                x_ticklabel_fs = None
            else:
                if kwargs['x_ticklabel_fs'] == "by default":
                    x_ticklabel_fs = None
                else:
                    x_ticklabel_fs = kwargs['x_ticklabel_fs']

            if "y_ticklabel_fs" not in kwarg_keys:
                y_ticklabel_fs = None
            else:
                if kwargs['y_ticklabel_fs'] == "by default":
                    y_ticklabel_fs = None
                else:
                    y_ticklabel_fs = kwargs['y_ticklabel_fs']

            if "x_label_fs" not in kwarg_keys:
                x_label_fs = None
            else:
                if kwargs['x_label_fs'] == "by default":
                    x_label_fs = None
                else:
                    x_label_fs = kwargs['x_label_fs']

            if "y_label_fs" not in kwarg_keys:
                y_label_fs = None
            else:
                if kwargs['y_label_fs'] == "by default":
                    y_label_fs = None
                else:
                    y_label_fs = kwargs['y_label_fs']

            if "title_fs" not in kwarg_keys:
                title_fs = None
            else:
                if kwargs['title_fs'] == "by default":
                    title_fs = None
                else:
                    title_fs = kwargs['title_fs']

            if "x_format" not in kwarg_keys:
                x_format = None
            else:
                if kwargs['x_format'] == "by default":
                    x_format = None
                else:
                    x_format = kwargs['x_format']

            if "y_format" not in kwarg_keys:
                y_format = None
            else:
                if kwargs['y_format'] == "by default":
                    y_format = None
                else:
                    y_format = kwargs['y_format']

            if "decoration_mark" not in kwarg_keys:
                decoration_mark = False
            else:
                if kwargs['decoration_mark'] == "by default":
                    decoration_mark = False
                else:
                    decoration_mark = kwargs['decoration_mark']

            if "marker_list" not in kwarg_keys:
                marker_list = None
            else:
                if kwargs['marker_list'] == "by default":
                    marker_list = None
                else:
                    marker_list = kwargs['marker_list']

            if "marker_label_list" not in kwarg_keys:
                marker_label_list = None
            else:
                if kwargs['marker_label_list'] == "by default":
                    marker_label_list = None
                else:
                    marker_label_list = kwargs['marker_label_list']

            if "marker" not in kwarg_keys:
                marker = None
            else:
                if kwargs['marker'] == "by default":
                    marker = None
                else:
                    marker = kwargs['marker']

            if "marker_size" not in kwarg_keys:
                marker_size = None
            else:
                if kwargs['marker_size'] == "by default":
                    marker_size = None
                else:
                    marker_size = kwargs['marker_size']

            if "xl_mark" not in kwarg_keys:
                xl_mark = False
            else:
                if kwargs['xl_mark'] == "by default":
                    xl_mark = False
                else:
                    xl_mark = kwargs['xl_mark']

            if "yl_mark" not in kwarg_keys:
                yl_mark = False
            else:
                if kwargs['yl_mark'] == "by default":
                    yl_mark = False
                else:
                    yl_mark = kwargs['yl_mark']

            if "line_color" not in kwarg_keys:
                line_color = None
            else:
                if kwargs['line_color'] == "by default":
                    line_color = None
                else:
                    line_color = kwargs['line_color']

            if "marker_color" not in kwarg_keys:
                marker_color = None
            else:
                if kwargs['marker_color'] == "by default":
                    marker_color = None
                else:
                    marker_color = kwargs['marker_color']

            if "marker_edge_width" not in kwarg_keys:
                marker_edge_width = None
            else:
                if kwargs['marker_edge_width'] == "by default":
                    marker_edge_width = None
                else:
                    marker_edge_width = kwargs['marker_edge_width']

            if "marker_face_color" not in kwarg_keys:
                marker_face_color = 'none'
            else:
                if kwargs['marker_face_color'] == "by default":
                    marker_face_color = 'none'
                else:
                    marker_face_color = kwargs['marker_face_color']

            if "legend_fs" not in kwarg_keys:
                legend_fs = 'none'
            else:
                if kwargs['legend_fs'] == "by default":
                    legend_fs = 'none'
                else:
                    legend_fs = kwargs['legend_fs']

            if "legend_loc" not in kwarg_keys:
                legend_loc = None
            else:
                if kwargs['legend_loc'] == "by default":
                    legend_loc = None
                else:
                    legend_loc = kwargs['legend_loc']

            canvas = draw(
                ax=kwargs['ax'],
                x=kwargs['x'],
                y=kwargs['y'],
                xl_mark=xl_mark,
                yl_mark=yl_mark,
                title=title,
                marker=marker,
                line_color=line_color,
                marker_color=marker_color,
                marker_size=marker_size,
                decoration_mark=decoration_mark,
                marker_list=marker_list,
                marker_label_list=marker_label_list,
                marker_edge_width=marker_edge_width,
                marker_face_color=marker_face_color,
                legend_loc=legend_loc,
                x_label=x_label,
                y_label=y_label,
                color=color,
                width=width,
                edgecolor=edgecolor,
                facecolor=facecolor,
                linewidth=linewidth,
                label=label,
                x_format=x_format,
                y_format=y_format,
                x_ticks=x_ticks,
                y_ticks=y_ticks,
                x_ticklabels=x_ticklabels,
                y_ticklabels=y_ticklabels,

                x_ticklabel_fs=x_ticklabel_fs,
                x_label_rotation=x_label_rotation,
                x_label_rotation_align=x_label_rotation_align,
                x_label_fs=x_label_fs,
                y_ticklabel_fs=y_ticklabel_fs,
                y_label_fs=y_label_fs,
                title_fs=title_fs,
                legend_fs=legend_fs,

                alpha=alpha,
            )
            params_all["canvas"] = canvas
            return params_all
        return config

    def bar_individual(self, **kwargs):
        canvas = kwargs["ax"].bar(
            x=kwargs["x"],
            height=kwargs["y"],
            width=kwargs["width"],
            color=kwargs["color"],
            label=kwargs["label"],
            edgecolor=kwargs["edgecolor"],
            facecolor=kwargs["facecolor"],
            linewidth=kwargs["linewidth"],
            alpha=kwargs["alpha"],
        )
        if kwargs["y_format"] == 'int':
            kwargs["ax"].yaxis.set_major_formatter(FuncFormatter(lambda y, _: '{:d}'.format(int(y))))
        elif kwargs["y_format"] == 'pct':
            kwargs["ax"].yaxis.set_major_formatter(FuncFormatter(lambda y, _: '{:.0%}'.format(y)))
        elif kwargs["y_format"] == 'exp':
            kwargs["ax"].yaxis.set_major_formatter(FuncFormatter(lambda y, _: '{:.2e}'.format(y)))
        else:
            kwargs["ax"].yaxis.set_major_formatter(FuncFormatter(lambda y, _: '{}'.format(y)))
        kwargs["ax"].spines['right'].set_color('none')
        kwargs["ax"].spines['top'].set_color('none')
        if kwargs["x_ticks"]:
            kwargs["ax"].set_xticks(kwargs["x"])
        if kwargs["y_ticks"]:
            kwargs["ax"].set_yticks(kwargs["y_ticks"])
        if kwargs["x_ticklabels"] is not None or kwargs["x_ticklabels"] == []:
            kwargs["ax"].set_xticklabels(kwargs["x_ticklabels"], rotation=15, ha='right', fontsize=kwargs["x_ticklabel_fs"])
        if kwargs["y_ticklabels"] is not None or kwargs["y_ticklabels"] == []:
            kwargs["ax"].set_yticklabels(kwargs["y_ticklabels"], fontsize=kwargs["y_ticklabel_fs"])
        if kwargs["xl_mark"]:
            kwargs["ax"].set_xlabel(kwargs["x_label"], fontsize=kwargs["x_label_fs"])
            kwargs['ax'].set_xticklabels(
                kwargs['x'],
                rotation=kwargs['x_label_rotation'],
                ha=kwargs['x_label_rotation_align'],
                fontsize=kwargs['x_ticklabel_fs'],
            )
        if kwargs["yl_mark"]:
            kwargs["ax"].set_ylabel(kwargs["y_label"], fontsize=kwargs["y_label_fs"])
        kwargs["ax"].set_title(kwargs["title"], fontsize=kwargs["title_fs"])
        return canvas

    def hist_grouped(self, **kwargs):
        # tc = plt.cm.tab20b(np.linspace(0, 1, len(y.keys()))).tolist()
        import seaborn as sns
        tc = sns.color_palette("Set2")
        for i, (label, y1) in enumerate(kwargs['y'].items()):
            kwargs['ax'].plot(
                kwargs['x'],
                y1,
                label=label,
                c=tc[i],
                alpha=kwargs['alpha'],
            )
        kwargs['ax'].spines['right'].set_color('none')
        kwargs['ax'].spines['top'].set_color('none')
        # if kwargs['x_ticks']:
        #     kwargs['ax'].set_xticks(kwargs['x_ticks'])
        # if kwargs['y_ticks']:
        #     kwargs['ax'].set_yticks(kwargs['y_ticks'])
        # if kwargs['x_ticklabels'] is not None or kwargs['x_ticklabels'] == []:
        #     kwargs['ax'].set_xticklabels(kwargs['x_ticklabels'], rotation=15, ha='right', fontsize=kwargs['x_ticklabel_fs'])
        # if kwargs['y_ticklabels'] is not None or kwargs['y_ticklabels'] == []:
        #     kwargs['ax'].set_yticklabels(kwargs['y_ticklabels'], fontsize=kwargs['y_ticklabel_fs'])
        if kwargs['xl_mark']:
            print(True)
            kwargs['ax'].set_xlabel(kwargs['x_label'], fontsize=kwargs['x_label_fs'])
            kwargs['ax'].set_xticks(kwargs['x_ticks'])
            kwargs['ax'].set_xticklabels(
                kwargs['x_ticklabels'],
                rotation=kwargs['x_label_rotation'],
                ha=kwargs['x_label_rotation_align'],
                fontsize=kwargs['x_ticklabel_fs'],
            )
        if kwargs['yl_mark']:
            kwargs['ax'].set_ylabel(kwargs['y_label'], fontsize=kwargs['y_label_fs'])
        kwargs['ax'].set_title(kwargs['title'], fontsize=kwargs['title_fs'])
        kwargs['ax'].legend(loc=kwargs['legend_loc'], ncol=1, fontsize=kwargs['legend_fs'])
        return

    def line(self, **kwargs):
        kwargs['ax'].plot(
            kwargs['x'],
            kwargs['y'],
            color=kwargs['line_color'],
            label=kwargs['label'],  # " ".join(ds_key.split("_"))
            linewidth=kwargs['linewidth'],
            marker=kwargs['marker'],
            markersize=kwargs['marker_size'],
            markeredgewidth=kwargs['marker_edge_width'],
            markerfacecolor=kwargs['marker_face_color'],
            # linestyle=line_styles[configs[ds_key]['position']],
        )
        kwargs['ax'].spines['right'].set_color('none')
        kwargs['ax'].spines['top'].set_color('none')
        if kwargs['decoration_mark']:
            kwargs['ax'].set_xlabel(kwargs['x_label'], fontsize=kwargs['x_label_fs'])
            kwargs['ax'].set_ylabel(kwargs['y_label'], fontsize=kwargs['y_label_fs'])
            if kwargs['title']:
                kwargs['ax'].set_title(kwargs['title'], fontsize=kwargs['title_fs'])
            kwargs['ax'].legend(loc=kwargs['legend_loc'], ncol=1, fontsize=kwargs['legend_fs'])
            kwargs['ax'].tick_params(axis='both', which='major', labelsize=14)
            kwargs['ax'].tick_params(axis='both', which='minor', labelsize=14)

        if kwargs["x_ticks"] and ["xl_mark"] is True:
            kwargs["ax"].set_xticks(kwargs["x_ticks"])
        if kwargs["y_ticks"] and ["xl_mark"] is True:
            kwargs["ax"].set_yticks(kwargs["y_ticks"])
        if kwargs["x_ticklabels"] is not None or kwargs["x_ticklabels"] == []:
            kwargs["ax"].set_xticklabels([], rotation=15, ha='right', fontsize=kwargs["x_ticklabel_fs"])
        if kwargs["y_ticklabels"] is not None or  kwargs["y_ticklabels"] == []:
            kwargs["ax"].set_yticklabels(kwargs["y_ticklabels"], fontsize=kwargs["y_ticklabel_fs"])
        if kwargs["xl_mark"]:
            kwargs["ax"].set_xlabel(kwargs["x_label"], fontsize=kwargs["x_label_fs"])
            kwargs["ax"].set_xticks(kwargs["x_ticks"])
            kwargs['ax'].set_xticklabels(
                kwargs['x_ticklabels'],
                rotation=kwargs['x_label_rotation'],
                ha=kwargs['x_label_rotation_align'],
                fontsize=kwargs['x_ticklabel_fs'],
            )
        if kwargs["yl_mark"]:
            kwargs["ax"].set_ylabel(kwargs["y_label"], fontsize=kwargs["y_label_fs"])
        kwargs["ax"].set_title(kwargs["title"], fontsize=kwargs["title_fs"])
        kwargs['ax'].legend(
            loc=kwargs['legend_loc'],
            ncol=1,
            fontsize=kwargs['legend_fs'],
            frameon=False,
        )

    def line_scatter(self, **kwargs):
        kwargs['ax'].plot(
            kwargs['x'],
            kwargs['y'],
            color=kwargs['line_color'],
            label=kwargs['label'],
            linewidth=kwargs['linewidth'],
            alpha=kwargs['alpha'],
            # marker="o" if configs[key]['position'] == "Cells" else "*",
            # markersize=5,
            # linestyle=line_styles[configs[key]['position']],
        )
        kwargs['ax'].scatter(
            kwargs['x'],
            kwargs['y'],
            color=kwargs['marker_color'],
            marker=kwargs['marker'],
            s=kwargs['marker_size'],
            facecolors='none',
            alpha=kwargs['alpha'],
        )
        if kwargs['decoration_mark']:
            print(True)
            kwargs['ax'].spines['right'].set_color('none')
            kwargs['ax'].spines['top'].set_color('none')
            kwargs["ax"].set_xticks(kwargs["x"])
            kwargs['ax'].set_xticklabels(
                kwargs['x'],
                rotation=kwargs['x_label_rotation'],
                ha=kwargs['x_label_rotation_align'],
                fontsize=kwargs['x_ticklabel_fs'],
            )

            kwargs['ax'].set_title(kwargs['title'], fontsize=kwargs['title_fs'])
            kwargs['ax'].set_xlabel(kwargs['x_label'], fontsize=kwargs['x_label_fs'])
            kwargs['ax'].set_ylabel(kwargs['y_label'], fontsize=kwargs['y_label_fs'])
            legend_main = kwargs['ax'].legend(
                loc='center left',
                ncol=1,
                bbox_to_anchor=(1, 0.5),
                frameon=False,
            )
            kwargs['ax'].add_artist(legend_main)
            import matplotlib.lines as mlines
            handles1 = [
                mlines.Line2D(
                    [],
                    [],
                    marker=marker,
                    markeredgecolor='k',
                    mfc='w',
                    ls='',
                ) for marker in kwargs['marker_list']
            ]
            legend_sub1 = kwargs['ax'].legend(
                handles1,
                kwargs['marker_label_list'],
                ncols=2,
                loc=kwargs['legend_loc'],
                fontsize=kwargs['legend_fs'],
            )
            kwargs['ax'].add_artist(legend_sub1)
            # handles2 = [mlines.Line2D([], [], linestyle=line_style, color='grey', mfc='w',) for line_style in [*line_styles.values()]]
            # ax[i].legend(handles2, [*line_styles.keys()], loc='lower right', ncols=3, fontsize=10)

            # handles = [mlines.Line2D([], [], marker=marker, mec='k', mfc='w', ls='') for marker in ['o', '^']]
            # ax[i].legend(handles, ['Radial', 'Transit'], loc='upper left', title="Detection")
        return


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.simu.Parameter import Parameter as params
from umiche.deduplicate.io.Stat import Stat as dedupstat
from umiche.plot.gadget.Transmitter import Transmitter as transmitter

from umiche.util.Reader import Reader as freader
from umiche.util.Console import Console


class DedupMultiple:

    def __init__(
            self,
            scenarios: Dict,
            methods: Dict,
            umi_gt_cnt: int = 50,
            param_fpn : str = None,
            verbose : bool = False,
    ):
        self.scenarios = scenarios
        self.methods = methods
        self.umi_gt_cnt = umi_gt_cnt
        self.param_fpn = param_fpn
        self.freader = freader()
        self.params = params(param_fpn=self.param_fpn)

        self.dedupstat = dedupstat(
            scenarios=self.scenarios,
            methods=self.methods,
            param_fpn=self.param_fpn,
        )

        self.df_dedup = self.dedupstat.df_dedup

        sns.set(font="Helvetica")
        sns.set_style("ticks")

        self.console = Console()
        self.console.verbose = verbose

    def line(
            self,
            num_img_row=2,
            num_img_col=3,
    ):
        print(self.df_dedup)
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(17, 9), sharey=False, sharex=False)
        palette = sns.color_palette("Paired")
        for n, (scenario, scenario_formal) in enumerate(self.scenarios.items()):
            self.df_sce = self.df_dedup[self.df_dedup['scenario'] == scenario_formal]
            for j, (method, method_formal) in enumerate(self.methods.items()):
                self.df_sce_met = self.df_sce[self.df_sce['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False

                self.line_gadget(
                    ax=ax[int(n / num_img_col), n % num_img_col],
                    x=self.df_sce_met.index,
                    y=self.df_sce_met['mean'].values,
                    line_color=palette[j+1],
                    label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    title='{}'.format(scenario_formal),
                    title_fs=16,
                    x_label=None, # 'Position'
                    y_label=r'($\frac{N_e-N_t}{N_t}$)',
                    x_label_rotation=20,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(self.df_sce_met.index.shape[0])).tolist(),
                    x_ticklabels=self.df_sce_met['metric'].values.tolist(),
                    x_ticklabel_fs=8,
                    x_label_fs=11,
                    y_label_fs=18,
                    legend_fs=11,
                    legend_loc=None,
                )
        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.06,
            left=0.06,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        plt.show()
        return

    @transmitter(type="line", task=None)
    def line_gadget(*args, **kwargs):
        return kwargs

    @transmitter(type="line_scatter", task=None)
    def line_scatter_gadget(*args, **kwargs):
        return kwargs


if __name__ == "__main__":
    from umiche.path import to

    p = DedupMultiple(
        scenarios={
            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'ampl_rates': 'Amplification rate',
            'umi_lens': 'UMI length',
            'seq_deps': 'Sequencing depth',
        },

        methods={
            # 'unique': 'Unique',
            # 'cluster': 'Cluster',
            # 'adjacency': 'Adjacency',
            'directional': 'Directional',
            'dbscan_seq_onehot': 'DBSCAN',
            'birch_seq_onehot': 'Birch',
            # 'aprop_seq_onehot': 'Affinity Propagation',
            'mcl': 'MCL',
            'mcl_val': 'MCL-val',
            'mcl_ed': 'MCL-ed',
        },

        param_fpn=to('data/params.yml'),
    )

    p.line()